""" Module providing a basic routine for plotting a shot overview """

# Common
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.gridspec import GridSpec

# tofu
try:
    from tofu.version import __version__
    import tofu.utils as utils
except Exception:
    from tofu.version import __version__
    from .. import utils as utils


__all__ = ['plot_shotoverview']

_fs = (12,6)
__github = 'https://github.com/ToFuProject/tofu/issues'
_wintit = 'tofu-%s        report issues / requests at %s'%(__version__, __github)
_dmargin = dict(left=0.04, right=0.99,
                bottom=0.07, top=0.93,
                wspace=0.25, hspace=0.12)
_fontsize = 8
_labelpad = 0
_dcol = {'Ip':'k', 'B':'b', 'Bt':'b',
         'PLH1':(1.,0.,0.),'PLH2':(1.,0.5,0.),
         'PIC1':'',
         'Prad':(1.,0.,1.),
         'q1rhot':(0.8,0.8,0.8),
         'Ax':(0.,1.,0.)}
_lct = [plt.cm.tab20.colors[ii] for ii in [0,2,4,1,3,5]]
_ntMax = 3


def plot_shotoverview(db, ntMax=_ntMax, indt=0, config=None, inct=[1,5],
                      dcol=None, lct=_lct, fmt_t='06.3f',
                      fs=None, dmargin=None, tit=None, wintit=None,
                      fontsize=_fontsize, labelpad=_labelpad,
                      sharet=True, sharey=True, shareRZ=True,
                      connect=True, draw=True):

    kh = _plot_shotoverview(db, ntMax=ntMax, indt=0, config=config, inct=inct,
                            dcol=dcol, lct=lct, fmt_t=fmt_t,
                            fs=fs, dmargin=dmargin, tit=tit, wintit=wintit,
                            fontsize=fontsize, labelpad=labelpad,
                            sharet=sharet, sharey=sharey, shareRZ=shareRZ,
                            connect=connect, draw=draw)
    return kh


######################################################
#       plot new
######################################################


def _plot_shotoverview_init(ns=1, sharet=True, sharey=True, shareRZ=True,
                            fontsize=_fontsize, fs=None,
                            wintit=None, dmargin=None):

    # Fromat inputs
    if fs is None:
        fs = _fs
    elif type(fs) is str and fs.lower()=='a4':
        fs = (11.7,8.3)
    if wintit is None:
        wintit = _wintit
    if dmargin is None:
        dmargin = _dmargin

    # Make figure and axes
    fig = plt.figure(figsize=fs)
    if wintit is not None:
        fig.canvas.set_window_title(wintit)
    axarr = GridSpec(ns, 3, **dmargin)

    laxt = [None for ii in range(0,ns)]
    laxc = [None for ii in range(0,ns)]
    for ii in range(0,ns):
        if ii == 0:
            laxt[ii] = fig.add_subplot(axarr[ii,:2])
            laxc[ii] = fig.add_subplot(axarr[ii,2])
            sht = laxt[0] if sharet else None
            shy = laxt[0] if sharey else None
            shRZ = laxc[0] if shareRZ else None
        else:
            laxt[ii] = fig.add_subplot(axarr[ii,:2], sharex=sht, sharey=shy)
            laxc[ii] = fig.add_subplot(axarr[ii,2], sharex=shRZ, sharey=shRZ)
            if not shareRZ:
                ax2.set_aspect('equal', adjustable='datalim')

    laxc[-1].set_xlabel(r'$R$ ($m$)')
    laxt[-1].set_xlabel(r'$t$ ($s$)', fontsize=fontsize)

    # datalim or box must be chosen for shared axis depending on matplotlib
    # version => let matplotlib decide until support for matplotlib 2.X.X is
    # stopped
    laxc[0].set_aspect('equal')#, adjustable='box')

    xtxt = laxc[0].get_position().bounds[0]
    dx = laxc[0].get_position().bounds[2]
    Ytxt, DY = np.sum(laxc[0].get_position().bounds[1::2]), 0.1
    axtxtt = fig.add_axes([xtxt, Ytxt, dx, DY], fc='None')

    # xtxt, Ytxt, dx, DY = 0.01, 0.98, 0.15, 0.02
    # axtxtg = fig.add_axes([xtxt, Ytxt, dx, DY], fc='None')

    # Dict
    dax = {'t':laxt,
           'cross':laxc,
           'txtt':[axtxtt]}
           #'txtg':[axtxtg] # not useful, one group only

    # Formatting
    for kk in dax.keys():
        for ii in range(0,len(dax[kk])):
            dax[kk][ii].tick_params(labelsize=fontsize)
            if 'txt' in kk:
                dax[kk][ii].patch.set_alpha(0.)
                for ss in ['left','right','bottom','top']:
                    dax[kk][ii].spines[ss].set_visible(False)
                dax[kk][ii].set_xticks([]), dax[kk][ii].set_yticks([])
                dax[kk][ii].set_xlim(0,1),  dax[kk][ii].set_ylim(0,1)
    return dax





def _plot_shotoverview(db, ntMax=_ntMax, indt=0, config=None, inct=[1,5],
                       dcol=None, lct=_lct, fmt_t='06.3f',
                       fs=None, dmargin=None, tit=None, wintit=None,
                       fontsize=_fontsize, labelpad=_labelpad,
                       sharet=True, sharey=True, shareRZ=True,
                       connect=True, draw=True):
    #########
    # Prepare
    #########
    fldict = dict(fontsize=fontsize, labelpad=labelpad)

    # Preformat
    if dcol is None:
        dcol = _dcol

    ls = sorted(list(db.keys()))
    ns = len(ls)
    lcol = ['k','b','r','m']

    # Find common time limits
    tlim = np.vstack([np.vstack([(np.nanmin(vv['t']), np.nanmax(vv['t']))
                                 if 't' in vv.keys() else (-np.inf,np.inf)
                                 for vv in db[ss].values()])
                      for ss in ls])
    tlim = (np.min(tlim),np.max(tlim))

    # Find common (R,Z) lims if config=None
    lEq = ['Ax','X','Sep','q1']
    if config is None:
        Anycross = False
        Rmin, Rmax = np.full((ns,),np.inf), np.full((ns,),-np.inf)
        Zmin, Zmax = np.full((ns,),np.inf), np.full((ns,),-np.inf)
        for ii in range(0,ns):
            for kk in set(db[ls[ii]].keys()).intersection(lEq):
                if db[ls[ii]][kk]['data2D'].ndim == 2:
                    Rmin[ii] = min(Rmin[ii],np.nanmin(db[ls[ii]][kk]['data2D'][:,0]))
                    Rmax[ii] = max(Rmax[ii],np.nanmax(db[ls[ii]][kk]['data2D'][:,0]))
                    Zmin[ii] = min(Zmin[ii],np.nanmin(db[ls[ii]][kk]['data2D'][:,1]))
                    Zmax[ii] = max(Zmax[ii],np.nanmax(db[ls[ii]][kk]['data2D'][:,1]))
                else:
                    Rmin[ii] = min(Rmin[ii],np.nanmin(db[ls[ii]][kk]['data2D'][:,0,:]))
                    Rmax[ii] = max(Rmax[ii],np.nanmax(db[ls[ii]][kk]['data2D'][:,0,:]))
                    Zmin[ii] = min(Zmin[ii],np.nanmin(db[ls[ii]][kk]['data2D'][:,1,:]))
                    Zmax[ii] = max(Zmax[ii],np.nanmax(db[ls[ii]][kk]['data2D'][:,1,:]))
                Anycross = True
        Rlim = (np.nanmin(Rmin),np.nanmax(Rmax))
        Zlim = (np.nanmin(Zmin),np.nanmax(Zmax))
        if Anycross is False:
            Rlim = (1,3)
            Zlim = (-1,-1)

    # time vectors and refs
    lt = [None for ss in ls]
    lidt = [0 for ss in ls]
    for ii in range(0,ns):
        for kk in set(db[ls[ii]].keys()).intersection(lEq):
            lt[ii] = db[ls[ii]][kk]['t']
            lidt[ii] = id(db[ls[ii]][kk]['t'])
            break
        else:
            for kk in set(db[ls[ii]].keys()).difference(lEq):
                lt[ii] = db[ls[ii]][kk]['t']
                lidt[ii] = id(db[ls[ii]][kk]['t'])
                break
            else:
                msg = "No reference time vector found for shot %s"%str(ls[ii])
                warnings.warn(msg)

    # dlextra id
    for ii in range(0,ns):
        for kk in set(db[ls[ii]].keys()).intersection(lEq):
            db[ls[ii]][kk]['id'] = id(db[ls[ii]][kk]['data2D'])


    ##############
    # Plot static
    ##############

    dax = _plot_shotoverview_init(ns=ns, sharet=sharet, sharey=sharey,
                                  shareRZ=shareRZ, fontsize=fontsize,
                                  fs=fs, wintit=wintit, dmargin=dmargin)
    fig = dax['t'][0].figure

    if tit is None:
        tit = r"overview of shots " + ', '.join(map('{0:05.0f}'.format,ls))
    fig.suptitle(tit)


    # Plot config and time traces
    for ii in range(0,ns):
        dd = db[ls[ii]]

        # config
        if config is not None:
            dax['cross'][ii] = config.plot(proj='cross', lax=dax['cross'][ii],
                                           element='P', dLeg=None, draw=False)

        # time traces
        for kk in set(dd.keys()).difference(lEq):
            if 'c' in dd[kk].keys():
                c = dd[kk]['c']
            else:
                c = dcol[kk]
            lab = dd[kk]['label'] + ' (%s)'%dd[kk]['units']
            dax['t'][ii].plot(dd[kk]['t'], dd[kk]['data'],
                              ls='-', lw=1., c=c, label=lab)
        kk = 'Ax'
        if kk in dd.keys():
            if 'c' in dd[kk].keys():
                c = dd[kk]['c']
            else:
                c = dcol[kk]
            x = db[ls[ii]][kk]['data2D'][:,0]
            y = db[ls[ii]][kk]['data2D'][:,1]
            dax['t'][ii].plot(lt[ii], x,
                              lw=1., ls='-', label=r'$R_{Ax}$ (m)')
            dax['t'][ii].plot(lt[ii], y,
                              lw=1., ls='-', label=r'$Z_{Ax}$ (m)')

    dax['t'][0].axhline(0., ls='--', lw=1., c='k')

    dax['t'][0].legend(bbox_to_anchor=(0.,1.01,1.,0.1), loc=3,
                       ncol=5, mode='expand', borderaxespad=0.,
                       prop={'size':fontsize})
    dax['t'][0].set_xlim(tlim)
    if config is None:
        try:        # DB
            dax['cross'][0].set_xlim(Rlim)
            dax['cross'][0].set_ylim(Zlim)
        except Exception as err:         # DB
            print(Rlim, Zlim)
            print(Rmin, Rmax)
            print(Zmin, Zmax)
            raise err
    for ii in range(0,ns):
        dax['t'][ii].set_ylabel('{0:05.0f} data'.format(ls[ii]), fontsize=fontsize)
    dax['cross'][-1].set_ylabel(r'$Z$ ($m$)', fontsize=fontsize)


    ##################
    # Interactivity dict
    ##################
    dgroup = {'time':   {'nMax':ntMax, 'key':'f1',
                         'defid':lidt[0], 'defax':dax['t'][0]}}

    # Group info (make dynamic in later versions ?)
    # msg = '  '.join(['%s: %s'%(v['key'],k) for k, v in dgroup.items()])
    # l0 = dax['txtg'][0].text(0., 0., msg,
                             # color='k', fontweight='bold',
                             # fontsize=6., ha='left', va='center')

    # dref
    dref = dict([(lidt[ii], {'group':'time', 'val':lt[ii], 'inc':inct})
                 for ii in range(0,ns)])

    # ddata
    ddat = {}
    for ii in range(0,ns):
        for kk in set(db[ls[ii]].keys()).intersection(lEq):
            ddat[db[ls[ii]][kk]['id']] = {'val':db[ls[ii]][kk]['data2D'],
                                      'refids':[lidt[ii]]}

    # dax
    lax_fix = dax['cross'] + dax['txtt']  # + dax['txtg']
    dax2 = dict([(dax['t'][ii], {'ref':{lidt[ii]:'x'}}) for ii in range(0,ns)])

    dobj = {}


    ##################
    # Populating dobj

    # One-axes time txt
    for jj in range(0,ntMax):
        l0 = dax['txtt'][0].text((0.5+jj)/ntMax, 0., r'',
                                 color='k', fontweight='bold',
                                 fontsize=fontsize,
                                 ha='left', va='bottom')
        dobj[l0] = {'dupdate':{'txt':{'id':lidt[0], 'lrid':[lidt[0]],
                                      'bstr':'{0:%s}'%fmt_t}},
                    'drefid':{lidt[0]:jj}}


    # Time-dependent
    nan2 = np.array([np.nan])
    for ii in range(0,ns):

        # time vlines
        for jj in range(0,ntMax):
            l0 = dax['t'][ii].axvline(np.nan,
                                      c=lct[jj], ls='-', lw=1.)
            dobj[l0] = {'dupdate':{'xdata':{'id':lidt[ii], 'lrid':[lidt[ii]]}},
                        'drefid':{lidt[ii]:jj}}

        # Eq
        for kk in set(db[ls[ii]].keys()).intersection(lEq):
            id_ = db[ls[ii]][kk]['id']
            for jj in range(0,ntMax):
                l0, = dax['cross'][ii].plot(nan2, nan2,
                                            ls='-', c=lct[jj], lw=1.)
                dobj[l0] = {'dupdate':{'data':{'id':id_, 'lrid':[lidt[ii]]}},
                            'drefid':{lidt[ii]:jj}}


    ##################
    # Instanciate KeyHandler
    can = fig.canvas
    can.draw()

    kh = utils.KeyHandler_mpl(can=can,
                              dgroup=dgroup, dref=dref, ddata=ddat,
                              dobj=dobj, dax=dax2, lax_fix=lax_fix,
                              groupinit='time', follow=True)

    if connect:
        kh.disconnect_old()
        kh.connect()
    if draw:
        fig.canvas.draw()
    return kh



#!/usr/bin/env python

# Built-in
import os
import argparse

# Common
import numpy as np

_save = True
_here = os.path.abspath(os.path.dirname(__file__))
_Exp, _Cls, _name = os.path.split(__file__)[1].split('_')[:3]
assert not any([any([ss in s for ss in ['Notes','.']])
               for s in [_Exp, _Cls, _name]])

def get_notes():
    """ By convention : D is a length of the element, d is a gap  """

    notes = {'DPhi':{}, 'dPhi':{}}
    # Toroidal width (mm, inner outer)
    notes['DPhi']['In'] = 25.508
    notes['DPhi']['Out'] = 32.908

    # Inter tiles distance (mm, uniform)
    notes['dl'] = 1.500

    # Poloidal/Radial total length (mm)
    notes['DL'] = 573.995

    # Number of tiles radially
    notes['nb'] = 2
    notes['nbPhi'] = 38*12

    # Radial length of a tile (mm)
    notes['Dl'] = np.array([316.495,256.000])

    # Vertical height of tiles (mm, uniform)
    notes['DZ'] = 25.538

    # Toroidal space between needles (mm, inner outer)
    notes['dPhi']['In'] = 0.666
    notes['dPhi']['Out'] = 0.666

    # (X,Z,Y) polygon of one needle (mm) !!!!!! (X,Z,Y)
    # 1 mm should be added towards Z>0 in the direction normal to the divertor's upper surface
    notes['sampleXZY'] = [[-440.221, -606.218, 1854.847],
                          [-440.163, -581.217, 1854.860],
                          [-440.748, -579.362, 1857.546],
                          [-506.714, -694.150, 2133.992],
                          [-506.951, -694.466, 2134.837],
                          [-510.087, -699.924, 2147.977],
                          [-510.336, -702.527, 2149.054],
                          [-508.309, -724.780, 2140.295],
                          [-508.995, -721.633, 2143.101],
                          [-510.684, -703.089, 2150.401],
                          [-511.270, -701.984, 2152.937],
                          [-514.021, -706.772, 2164.465],
                          [-514.742, -708.491, 2167.329],
                          [-565.491, -796.886, 2380.007],
                          [-565.707, -799.489, 2381.092],
                          [-563.726, -821.241, 2372.530]]
    notes['sampleXZY'] = np.array(notes['sampleXZY'])

    for kk in notes.keys():
        if type(notes[kk]) is dict:
            notes[kk]['In'] = notes[kk]['In']*1.e-3
            notes[kk]['Out'] = notes[kk]['Out']*1.e-3
        elif not 'nb' in kk:
            notes[kk] = notes[kk]*1.e-3
    return notes


def _get_inter(D0,u0,D1,u1):
    k = -np.cross(D0-D1,u1)/np.cross(u0,u1)
    return D0 + k*u0

def make_Poly(save=_save, path=_here):
    notes = get_notes()

    R = np.hypot(notes['sampleXZY'][:,0],notes['sampleXZY'][:,2])
    Z = notes['sampleXZY'][:,1]

    d = np.sqrt(np.diff(R)**2+np.diff(Z)**2)
    indup = np.argmax(d)
    e1 = np.array([R[indup+1]-R[indup], Z[indup+1]-Z[indup]])
    e1 = e1/np.linalg.norm(e1)
    e2 = np.r_[-e1[1], e1[0]]

    P0 = (np.array([R[0],Z[0]])-0.01*e2)[:,np.newaxis]
    PEnd = (np.array([R[-1],Z[-1]])-0.01*e2)[:,np.newaxis]
    Poly = np.array([R[1:-1],Z[1:-1]])
    Poly = np.concatenate((P0,Poly,PEnd),axis=1)
    Poly0 = Poly[:,[0,2,-3,-1]]

    # Making Poly1
    D0 = Poly0[:,1]
    u0 = Poly0[:,-2]-D0
    k0 = -np.sum((D0-Poly[:,5])*u0)/np.linalg.norm(u0)**2
    k1 = -np.sum((D0-Poly[:,10])*u0)/np.linalg.norm(u0)**2
    P0 = (D0 + k0*u0)[:,np.newaxis]
    P1 = (D0 + k1*u0)[:,np.newaxis]
    D0, D1 = Poly[:,5], Poly[:,10]
    u0, u1 = Poly[:,6]-D0, Poly[:,9]-D1
    PI = _get_inter(D0,u0,D1,u1)[:,np.newaxis]
    Poly1 = np.concatenate((Poly0[:,:2], P0,PI,P1, Poly0[:,2:]),axis=1)

    if save:
        cstr = '%s_%s_%s'%(_Exp,_Cls,_name)
        pathfilext = os.path.join(path, cstr+'_V0.txt')
        np.savetxt(pathfilext, Poly0.T)
        pathfilext = os.path.join(path, cstr+'_V1.txt')
        np.savetxt(pathfilext, Poly1.T)
        pathfilext = os.path.join(path, cstr+'_V2.txt')
        np.savetxt(pathfilext, Poly.T)
    return Poly0, Poly1, Poly, notes


if __name__=='__main__':

    # Parse input arguments
    msg = 'Launch creation of polygons txt from bash'
    parser = argparse.ArgumentParser(description = msg)

    parser.add_argument('-save', type=bool, help='save ?', default=_save)
    parser.add_argument('-path', type=str, help='saving path ?', default=_here)

    args = parser.parse_args()

    # Call wrapper function
    make_Poly(save=args.save, path=args.path)



"""
This module contains tests for tofu.geom in its structured version
"""

# External modules
import os
import numpy as np
import matplotlib.pyplot as plt


# Nose-specific
from nose import with_setup # optional


# Importing package tofu.geom


#Root = tfpf.Find_Rootpath()


#######################################################
#
#     Setup and Teardown
#
#######################################################

def setup_module(module):
    print ("") # this is to get a newline after the dots
    #print ("setup_module before anything in this file")

def teardown_module(module):
    #os.remove(VesTor.Id.SavePath + VesTor.Id.SaveName + '.npz')
    #os.remove(VesLin.Id.SavePath + VesLin.Id.SaveName + '.npz')
    #print ("teardown_module after everything in this file")
    #print ("") # this is to get a newline
    pass


#def my_setup_function():
#    print ("my_setup_function")

#def my_teardown_function():
#    print ("my_teardown_function")

#@with_setup(my_setup_function, my_teardown_function)
#def test_numbers_3_4():
#    print 'test_numbers_3_4  <============================ actual test code'
#    assert multiply(3,4) == 12

#@with_setup(my_setup_function, my_teardown_function)
#def test_strings_a_3():
#    print 'test_strings_a_3  <============================ actual test code'
#    assert multiply('a',3) == 'aaa'



# -*- coding: utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
# Also if needed: retab
'''
    Magnetic field line tracing
'''
# Standard python modules
from __future__ import (unicode_literals, absolute_import,  \
                        print_function, division)
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import axes3d, Axes3D
import numpy as np
import os
#import re
import scipy.integrate as spode
import scipy.interpolate as interpolate
import scipy.spatial
#import warnings
#import sys

# Local modules
import imas
try:
    import pywed as pw
except ImportError as err:
    print(err)

# Project modules
try:
    from tofu.mag.mag_ripple.mag_ripple import mag_ripple
except Exception:
    from mag.mag_ripple.mag_ripple import mag_ripple
#try:
#    from equimap import interp_quantity
#except ImportError as err:
#    print(err)

__all__ = ['MagFieldLines']

class MagFieldLines:
    '''
    Compute magnetic field lines

    Parameters
    ----------
    shot : int
        Shot number
    run : run number, optional (default=0)
    occ : occurrence number, optional (default=0)
    user : user name, optional (default=imas_public)
    machine : machine name, optional (default=west)

    Methods
    -------
    trace_mline(init_points_array, time_array, direction='FWD')

    In other words:
    trace_mline([[r_1, r_2, ...], [phi_1, phi_2, ...], [z_1, z_2, ...]], [t1, t2, ...], direction='FWD')
        returns list of list of traces with shape (times, init_points)
    plot_trace(trace[0][0])
        2D plots of traced magnetic field line
    plot_trace_3D(trace[0][0])
        3D plot of traced magnetic field line

    Examples
    --------
    For shot 54095
    Compute linear interpolation B_r, B_tor, B_z at r=2, phi=0, z=0
    and r, z, phi of magnetic line starting at r=2.7, phi=0, z=0
    >>> test = MagFieldLines(54095)

    Call method with specific time (t=34 seconds in this case)
    >>> test.trace_mline([2.7, 0., 0.], 34)[0][0]['r'][10]
    2.700041665466963
    >>> test.trace_mline([2.7, 0., 0.], 34)[0][0]['p'][10]
    -0.0037010956581712776
    >>> test.trace_mline([2.7, 0., 0.], 34)[0][0]['z'][10]
    0.000373472578674799

    Call method for list of points (2 items and t=34 seconds)
    >>> test.trace_mline([[2.7, 2.8], [0., 0.], [0., 0.1*np.pi]], 34)[0][0]['r'][10]
    2.700041665466963
    >>> test.trace_mline([[2.7, 2.8], [0., 0.], [0., 0.1*np.pi]], 34)[0][0]['p'][10]
    -0.0037010956581712776
    >>> test.trace_mline([[2.7, 2.8], [0., 0.], [0., 0.1*np.pi]], 34)[0][0]['z'][10]
    0.000373472578674799

    For interpolation of B magnetic field
    >>> test.b_field_interp(2, 0, 0)
    (0.006836488559673934, -4.547017391384852, -0.14305889263210517)
    '''
    def __init__(self, shot, run=0, occ=0, user='imas_public', machine='west'):

        # Check if shot exists
        run_number = '{:04d}'.format(run)
        shot_file  = os.path.expanduser('~' + user + '/public/imasdb/' + machine + \
                                        '/3/0/' + 'ids_' + str(shot) + run_number + \
                                        '.datafile')
        if (not os.path.isfile(shot_file)):
            raise FileNotFoundError('IMAS file does not exist')

        # Parameters
        self.wall_ck=False

        # Get equilibrium
        idd = imas.ids(shot, run)
        idd.open_env(user, machine, '3')
        idd.equilibrium.get()
        self.equi = idd.equilibrium

        # Check code.output_flag for data validity
        if (np.any(np.isnan(self.equi.code.output_flag))):
            self.mask = np.full(len(self.equi.time), True, dtype=bool)
        else:
            self.mask = np.asarray(self.equi.code.output_flag) >= 0

        # Get Itor (current of toroidal coils, coils that produce the toroidal field)
        self.itor, self.t_itor = pw.tsbase(shot, 'gmag_itor', nargout=2)
        self.t_ignitron        = pw.tsmat(shot, 'IGNITRON|1')
        self.t_itor           += self.t_ignitron[0]

        self.equiDict = {}

        # Declaration of arrays 2d from equilibrium IDS
        equi_grid = idd.equilibrium.grids_ggd[0].grid[0]
        NbrPoints = len(equi_grid.space[0].objects_per_dimension[0].object)
        self.equiDict['r'] = np.full(NbrPoints, np.nan)
        self.equiDict['z'] = np.full(NbrPoints, np.nan)
        for ii in range(NbrPoints):
            self.equiDict['r'][ii] = equi_grid.space[0].objects_per_dimension[0]. \
                                object[ii].geometry[0]
            self.equiDict['z'][ii] = equi_grid.space[0].objects_per_dimension[0]. \
                                object[ii].geometry[1]

        self.equiDict['b_field_r']   = np.full((len(self.equi.time), NbrPoints), np.nan)
        self.equiDict['b_field_z']   = np.full((len(self.equi.time), NbrPoints), np.nan)
        self.equiDict['b_field_tor'] = np.full((len(self.equi.time), NbrPoints), np.nan)
        self.equiDict['boundary_r']  = [None]*len(self.equi.time)
        self.equiDict['boundary_z']  = [None]*len(self.equi.time)
        for ii in range(len(self.equi.time)):
            equi_slice  = self.equi.time_slice[ii]
            self.equiDict['b_field_r'][ii]   = equi_slice.ggd[0].b_field_r[0].values
            self.equiDict['b_field_z'][ii]   = equi_slice.ggd[0].b_field_z[0].values
            self.equiDict['b_field_tor'][ii] = equi_slice.ggd[0].b_field_tor[0].values
            self.equiDict['boundary_r'][ii]  = equi_slice.boundary.outline.r
            self.equiDict['boundary_z'][ii]  = equi_slice.boundary.outline.z

        self.points   = np.vstack((self.equiDict['r'], self.equiDict['z'])).transpose()
        self.delaunay = scipy.spatial.Delaunay(self.points)

        # Time interpolation
        self.f_intp_br = interpolate.interp1d(self.equi.time[self.mask], \
                           self.equiDict['b_field_r'][self.mask, :], axis=0, \
                           bounds_error=False)
        self.f_intp_bt = interpolate.interp1d(self.equi.time[self.mask], \
                           self.equiDict['b_field_tor'][self.mask, :], axis=0, \
                           bounds_error=False)
        self.f_intp_bz = interpolate.interp1d(self.equi.time[self.mask], \
                           self.equiDict['b_field_z'][self.mask, :], axis=0, \
                           bounds_error=False)

    def trace_mline(self, init_state, time, direction='FWD', \
                    length_line=None, stp=None, ripple=True):
        '''
        Traces the field line given a starting point.
        Integration step defined by stp and maximum length of the field line
        defined by s.
        Collision with the wall stops integration.

        input:
            - init_state [[r_1, r_2, ...], [phi_1, phi_2, ...], [z_1, z_2, ...]] :
                  the coordinates of the starting points (list or numpy array)
                  OR list of lists OR numpy 2D array
            - time : array of times requested (list or numpy array)
            - direction : direction of integration 'FWD' forward or 'REV' reverse
              (string)
            - ripple : take into account magnetic ripple or not (boolean)

        output:
            returns a dictionary containing:
            - r  : radial coordinate (np.array)
            - p  : toridal coordinate (np.array)
            - z  : vertical coordinate (np.array)
            - x  : x coordinate (np.array)
            - y  : y coordinate (np.array)
            - cp : collision point with the wall (list)

        '''
        # Step for the integration
        if (stp is None):
            stp = 0.001

        # Length of the field line
        if (length_line is None):
            s = 100
        else:
            s = length_line

        ds  = np.linspace(0, s, int(s/stp))

        init_state = np.squeeze(np.asarray(init_state))
        if (init_state.ndim < 2):
            init_state = init_state[:, np.newaxis]

        self.ar_time = np.atleast_1d(np.squeeze(np.asarray([time])))

        br_intp_t = np.atleast_2d(np.squeeze(self.f_intp_br(self.ar_time)))
        bt_intp_t = np.atleast_2d(np.squeeze(self.f_intp_bt(self.ar_time)))
        bz_intp_t = np.atleast_2d(np.squeeze(self.f_intp_bz(self.ar_time)))

        # !!!!!!!!!!!!!!!!!!!!!
        # HARD CODED CORRECTION
        if (np.nanmean(bt_intp_t) > 0):
            bt_intp_t *= -1
        # !!!!!!!!!!!!!!!!!!!!!

        # Interpolate current
        itor_intp_t_vect = np.interp(self.ar_time, self.t_itor[:, 0], \
                                     self.itor[:, 0])
        b0_intp_t_vect   = np.interp(self.ar_time, self.equi.time[self.mask], \
                              self.equi.vacuum_toroidal_field.b0[self.mask])

        outMagLine = []

        for ii in range(self.ar_time.size):
            self.br_lin_intp = \
              interpolate.LinearNDInterpolator(self.delaunay, br_intp_t[ii])
            self.bt_lin_intp = \
              interpolate.LinearNDInterpolator(self.delaunay, bt_intp_t[ii])
            self.bz_lin_intp = \
              interpolate.LinearNDInterpolator(self.delaunay, bz_intp_t[ii])

            # Interpolated current and b0
            self.itor_intp_t = itor_intp_t_vect[ii]
            self.b0_intp_t   = b0_intp_t_vect[ii]

            out = []

            for jj in range(init_state.shape[1]):
                out_prime = self.integrate_solve_ivp(init_state[:, jj], \
                                                     direction, s, ds, ripple)
                out_prime['init_point'] = init_state[:, jj]
                out_prime['time']       = self.ar_time[ii]
                # Return list of dict
                out.append(out_prime)

            outMagLine.append(out)

        return outMagLine

    def integrate_solve_ivp(self, init_state, direction, s, ds, ripple=True):
        '''
        Integration step defined by stp and maximum length of the field line
        defined by s.
        Collision with the wall stops integration.

        input:
            - init_state [r, phi, z] : the coordinates of the starting point (list)
            - direction : direction of integration 'FWD' forward or 'REV' reverse
              (string)
            - s : field line maximum length
            - ds : vector of steps
            - ripple : boolean, take into account magnetic ripple or not

        output:
            returns a dictionary containing:
            - r  : radial coordinate (np.array)
            - p  : toridal coordinate (np.array)
            - z  : vertical coordinate (np.array)
            - x  : x coordinate (np.array)
            - y  : y coordinate (np.array)
            - cp : collision point with the wall (list)
        '''
        if (ripple):
            if direction == 'FWD':
                sol = spode.solve_ivp(self.mfld3dcylfwd, [0, s], init_state,
                                      method='RK23', t_eval=ds,
                                      events=self.hit_wall_circ)
            elif direction == 'REV':
                sol = spode.solve_ivp(self.mfld3dcylrev, [0, s], init_state,
                                      method='RK23', t_eval=ds,
                                      events=self.hit_wall_circ)
        else:
            if direction == 'FWD':
                sol = spode.solve_ivp(self.mfld3dcylfwd_no_ripple,
                                      [0, s], init_state,
                                      method='RK23', t_eval=ds,
                                      events=self.hit_wall_circ)
            elif direction == 'REV':
                sol = spode.solve_ivp(self.mfld3dcylrev_no_ripple,
                                      [0, s], init_state,
                                      method='RK23', t_eval=ds,
                                      events=self.hit_wall_circ)
        sgf = sol.t
        rgf = sol.y[0]
        pgf = sol.y[1]
        xgf = rgf*np.cos(pgf)
        ygf = rgf*np.sin(pgf)
        zgf = sol.y[2]

        if len(sgf)<len(ds):
            colpt = [rgf[-1], pgf[-1], zgf[-1]]
        else:
            colpt = []

        return {'s':  sgf,
                'r':  rgf,
                'p':  pgf,
                'z':  zgf,
                'x':  xgf,
                'y':  ygf,
                'cp': colpt}

    def mfld3dcylfwd(self, s, state):
        '''
        Returns the right end side of the field line system of equations in
        cyclindrical (R, Phi, Z) coord.
        This is the case for forward integration.
        '''
        R, P, Z = state
        Br, Bt, Bz = self.b_field_interp(R, P, Z)
        #Br=self.fBp_r(R,Z)[0]+self.fBt_r(P,self.ftheta(R,Z),R)
        #Bt=self.fBt_t(P,self.ftheta(R,Z),R)
        #Bz=self.fBp_z(R,Z)[0]
        B = np.sqrt(Br*Br + Bz*Bz + Bt*Bt)
        d_R = Br/B
        d_P = Bt/B*1/R
        d_Z = Bz/B
        return [d_R, d_P, d_Z]

    def mfld3dcylrev(self, s, state):
        '''
        Returns the right end side of the field line system of equations in
        cyclindrical (R, Phi, Z) coord.
        This is the case for backward integration.
        '''
        R, P, Z = state
        Br, Bt, Bz = self.b_field_interp(R, P, Z)
        # Br=self.fBp_r(R,Z)[0]+self.fBt_r(P,self.ftheta(R,Z),R)
        # Bt=self.fBt_t(P,self.ftheta(R,Z),R)
        # Bz=self.fBp_z(R,Z)[0]
        B = np.sqrt(Br*Br + Bz*Bz + Bt*Bt)
        d_R = -Br/B
        d_P = -Bt/B*1/R
        d_Z = -Bz/B
        return [d_R, d_P, d_Z]

    def mfld3dcylfwd_no_ripple(self, s, state):
        '''
        Returns the right end side of the field line system of equations in
        cyclindrical (R, Phi, Z) coord.
        This is the case for forward integration.
        '''
        R, P, Z = state
        Br, Bt, Bz = self.b_field_interp_no_ripple(R, P, Z)
        # Br=self.fBp_r(R,Z)[0]+self.fBt_r(P,self.ftheta(R,Z),R)
        # Bt=self.fBt_t(P,self.ftheta(R,Z),R)
        # Bz=self.fBp_z(R,Z)[0]
        B = np.sqrt(Br*Br + Bz*Bz + Bt*Bt)
        d_R = Br/B
        d_P = Bt/B*1/R
        d_Z = Bz/B
        return [d_R, d_P, d_Z]

    def mfld3dcylrev_no_ripple(self, s, state):
        '''
        Returns the right end side of the field line system of equations in
        cyclindrical (R, Phi, Z) coord.
        This is the case for backward integration.
        '''
        R, P, Z = state
        Br, Bt, Bz = self.b_field_interp_no_ripple(R, P, Z)
        # Br=self.fBp_r(R,Z)[0]+self.fBt_r(P,self.ftheta(R,Z),R)
        # Bt=self.fBt_t(P,self.ftheta(R,Z),R)
        # Bz=self.fBp_z(R,Z)[0]
        B = np.sqrt(Br*Br + Bz*Bz + Bt*Bt)
        d_R = -Br/B
        d_P = -Bt/B*1/R
        d_Z = -Bz/B
        return [d_R, d_P, d_Z]

    def hit_wall_circ(self, s, state):
        '''
        return 0 when hit wall.
        With wall_ck:
        - False => The wall is a simple circular torus with minor radius rw
            centered at (Rc,Zc), thus with major radius Rc.
        - True => check collision with wall boundary given in the Equilibrium
            mat file
        '''
        R, P, Z = state
        # if (np.abs(Z) <= 0.5 and R > 3.01
        #     and np.deg2rad(89.5) <= P <= np.deg2rad(90.5)):
        #     return 0
        # if (np.abs(Z) <= 0.5 and R > 3.01
        #     and np.deg2rad(179.5) <= P <= np.deg2rad(180.5)):
        #     print('Got 2')
        #     return 0
        # elif (np.abs(Z) <= 0.5 and R > 3.01
        #       and np.deg2rad(269.5) <= P <= np.deg2rad(270.5)):
        #     print('Got 3')
        #     return 0
        if self.wall_ck==False:
            Rc=2.460
            Zc=0.0
            rw=0.950
            return (R-Rc)**2+(Z-Zc)**2-rw**2
        elif self.wall_ck and Z>=0:
            zwall=self.fwall_up(R)
            return Z-zwall
        elif self.wall_ck and Z<0:
            zwall=self.fwall_dw(R)
            return zwall-Z

    hit_wall_circ.terminal=True

    def b_field_interp(self, R, Phi, Z):
        ''' Linear interpolation of B vector components at R, Phi, Z positions
        '''
        interp_points = np.vstack((R, Z)).transpose()

        br_intp = self.br_lin_intp.__call__(interp_points)
        bt_intp = self.bt_lin_intp.__call__(interp_points)
        bz_intp = self.bz_lin_intp.__call__(interp_points)

        # Compute magnetic field for given Phi
        br_ripple, bt_ripple, bz_ripple = mag_ripple(R, Phi, \
                                                     Z, self.itor_intp_t)
        # Compute reference vaccuum magnetic field
        self.bt_vac = self.equi.vacuum_toroidal_field.r0*self.b0_intp_t / R

        br_intp -= br_ripple[0]
        bt_intp -= (bt_ripple[0] - np.abs(self.bt_vac))
        bz_intp -= bz_ripple[0]

        return br_intp[0], bt_intp[0], bz_intp[0]

    def b_field_interp_no_ripple(self, R, Phi, Z):
        ''' Linear interpolation of B vector components at R, Phi, Z positions
        '''
        interp_points = np.vstack((R, Z)).transpose()

        br_intp = self.br_lin_intp.__call__(interp_points)
        bt_intp = self.bt_lin_intp.__call__(interp_points)
        bz_intp = self.bz_lin_intp.__call__(interp_points)

        return br_intp[0], bt_intp[0], bz_intp[0]

    def plot_trace(self, trace):
        '''
        Plots a summary of the magnetic field line trace.
        It contains a graph of the radial, vertical and toroidal corrdinates followed by
        the vertical projection (w/o and with the wall) and the toroidal projection.

        A black dot represents the starting point of the field line.
        The end of the trace is indicated by:
        - a red dot in case of collision with the wall
        - a black square otherwise

        input:
            - trace : the magneticl field line data (dictionary)
        '''
        sgf   = trace['s']
        rgf   = trace['r']
        zgf   = trace['z']
        pgf   = trace['p']
        xgf   = trace['x']
        ygf   = trace['y']
        colpt = trace['cp']

        ind_time = np.argmin(np.abs(self.equi.time - trace['time']))

        plt.figure(figsize=[12,8])

        plt.subplot(231)
        plt.plot(sgf, rgf)
        plt.grid()
        plt.ylabel('R [m]'); plt.xlabel('s [m]'); plt.title('Radial coord.')

        plt.subplot(232)
        plt.plot(sgf, zgf)
        plt.grid()
        plt.ylabel('Z [m]'); plt.xlabel('s [m]'); plt.title('Vertical coord.')

        plt.subplot(233)
        plt.plot(sgf, np.rad2deg(pgf))
        plt.grid()
        plt.ylabel('$\phi$ [deg]'); plt.xlabel('s [m]'); plt.title('Toroidal coord.')

        plt.subplot(234)
        plt.plot(rgf, zgf)
        plt.plot(rgf[0], zgf[0], marker='o', markersize=3, color="black")
        if not colpt:
            plt.plot(rgf[-1], zgf[-1], marker='s', markersize=5, color="black")
        else:
            plt.plot(rgf[-1], zgf[-1], marker='o', markersize=5, color="red")
        plt.axis('equal'); plt.xlabel('R [m]'); plt.ylabel('Z [m]')
        plt.plot(self.equiDict['boundary_r'][ind_time], \
                 self.equiDict['boundary_z'][ind_time])
        plt.title('Vertical projection')
        plt.grid()

        plt.subplot(235)
        plt.plot(rgf, zgf)
        #plt.plot(rwall,zwall)
        plt.plot(rgf[0], zgf[0], marker='o', markersize=3, color="black")
        if not colpt:
            plt.plot(rgf[-1], zgf[-1], marker='s', markersize=5, color="black")
        else:
            plt.plot(rgf[-1], zgf[-1], marker='o', markersize=5, color="red")
        plt.axis('equal'); plt.xlabel('R [m]'); plt.ylabel('Z [m]')
        plt.title('Vertical projection')

        plt.subplot(236)
        plt.plot(xgf, ygf)
        #plt.plot(ra*np.cos(pa),ra*np.sin(pa),'b:')
        plt.plot(xgf[0], ygf[0], marker='o', markersize=3, color="black")
        if not colpt:
            plt.plot(xgf[-1], ygf[-1], marker='s', markersize=5, color="black")
        else:
            plt.plot(xgf[-1], ygf[-1], marker='o', markersize=5, color="red")
        plt.axis('equal'); plt.xlabel('x [m]'); plt.ylabel('y [m]')
        plt.title('Toroidal projection')

        plt.tight_layout()
        plt.show()

    def plot_trace_3D(self, trace):
        '''
        Plots the magnetic field line trace in a 3D projection.

        The outline of the antenna is shown by black solid ines.
        A black dot represents the starting point of the field line.
        The end of the trace is indicated by:
        - a red dot in case of collision with the wall
        - a black square otherwise

        input:
            - trace : the magneticl field line data (dictionary)
        '''
        sgf   = trace['s']
        rgf   = trace['r']
        zgf   = trace['z']
        pgf   = trace['p']
        xgf   = trace['x']
        ygf   = trace['y']
        colpt = trace['cp']

        fig = plt.figure(figsize=[10, 10])
        ax = Axes3D(fig)
        #ax = fig.gca(projection='3d')
        #ax.set_aspect('equal')
        ax.plot(xgf, ygf, zgf)
        #ax.plot(ra[0]*np.cos(pa),ra[0]*np.sin(pa),za[0],'k')
        #ax.plot(ra*np.cos(pa[0]),ra*np.sin(pa[0]),za,'k')
        #ax.plot(ra[-1]*np.cos(pa),ra[-1]*np.sin(pa),za[-1],'k')
        #ax.plot(ra*np.cos(pa[-1]),ra*np.sin(pa[-1]),za,'k')
        ax.plot(3.5*np.cos(np.linspace(0, 2*np.pi, 36)), \
                3.5*np.sin(np.linspace(0, 2*np.pi, 36)), 0, 'k:')
        # Used to create the fake bounding box
        max_range = np.array([xgf.max()-xgf.min(), ygf.max()-ygf.min(), zgf.max()-zgf.min()]).max()
        Xb = 0.5*max_range*np.mgrid[-1:2:2,-1:2:2,-1:2:2][0].flatten() + 0.5*(xgf.max()+xgf.min())
        Yb = 0.5*max_range*np.mgrid[-1:2:2,-1:2:2,-1:2:2][1].flatten() + 0.5*(ygf.max()+ygf.min())
        Zb = 0.5*max_range*np.mgrid[-1:2:2,-1:2:2,-1:2:2][2].flatten() + 0.5*(zgf.max()+zgf.min())
        # Comment or uncomment following both lines to test the fake bounding box:
        for xb, yb, zb in zip(Xb, Yb, Zb):
            ax.plot([xb], [yb], [zb], 'w')
        ax.plot([xgf[0]], [ygf[0]], [zgf[0]], marker='o', markersize=3, color="black")
        if not colpt:
            ax.plot([xgf[-1]], [ygf[-1]], [zgf[-1]], marker='s', markersize=5, color="black")
        else:
            ax.plot([xgf[-1]], [ygf[-1]], [zgf[-1]], marker='o', markersize=5, color="red")
        ax.set_xlabel('x [m]'); ax.set_ylabel('y [m]'); ax.set_zlabel('z [m]')
        plt.title('Magnetic field line trace')
        plt.show()

    def plot_poincare(self, trace_plt, tor_angle=0):
        '''
        Plots Poincare plot at given toroidal angle

        input:
            - trace_plt : the magneticl field line data for several points
                          one single time (dictionary)
        '''

        fig, ax = plt.subplots()

        for ii in range(len(trace_plt)):
            loc_cross = trace_plt[ii]['p'] % (tor_angle + 2*np.pi)
            mask_cross = np.r_[False, loc_cross[1:]
                               < loc_cross[:-1]] & np.r_[loc_cross[:-1]
                                                         < loc_cross[1:],
                                                         False]
            ax.plot(trace_plt[ii]['r'][mask_cross],
                    trace_plt[ii]['z'][mask_cross], 'o', markersize=3)

        ax.set_aspect('equal')
        ax.set_xlabel('R [m]')
        ax.set_ylabel('Z [m]')




# Built-in
import os

# Common
import numpy as np


# #############################################################################
#                   Triangular meshes
# #############################################################################


def tri_checkformat_NodesFaces(nodes, indfaces, ids=None):

    # Check mesh type
    if indfaces.shape[1] == 3:
        mtype = 'tri'
    elif indfaces.shape[1] == 4:
        mtype = 'quad'
    else:
        msg = ("Mesh seems to be neither triangular nor quadrilateral\n"
               + "  => unrecognized mesh type, not implemented yet")
        raise Exception(msg)

    # Check indexing !!!
    indmax = int(np.nanmax(indfaces))
    if indmax == nodes.shape[0]:
        indfaces = indfaces - 1
    elif indmax > nodes.shape[0]:
        msg = ("There seems to be an indexing error\n"
               + "\t- np.max(indfaces) = {}".format(indmax)
               + "\t- nodes.shape[0] = {}".format(nodes.shape[0]))
        raise Exception(msg)

    # Check for duplicates
    nnodes = nodes.shape[0]
    nfaces = indfaces.shape[0]
    nodesu, indnodesu = np.unique(nodes, axis=0, return_index=True)
    facesu, indfacesu = np.unique(indfaces, axis=0, return_index=True)
    facesuu = np.unique(facesu)
    lc = [nodesu.shape[0] != nnodes,
          facesu.shape[0] != nfaces,
          facesuu.size != nnodes or np.any(facesuu != np.arange(0, nnodes))]
    if any(lc):
        msg = "Non-valid mesh in {}:\n".format(ids)
        if lc[0]:
            noddup = [ii for ii in range(0, nnodes) if ii not in indnodesu]
            msg += ("  Duplicate nodes: {}\n".format(nnodes - nodesu.shape[0])
                    + "\t- nodes.shape: {}\n".format(nodes.shape)
                    + "\t- unique nodes.shape: {}\n".format(nodesu.shape)
                    + "\t- duplicate nodes indices: {}\n".format(noddup))
        if lc[1]:
            dupf = [ii for ii in range(0, nfaces) if ii not in indfacesu]
            msg += ("  Duplicate faces: {}\n".format(nfaces - facesu.shape[0])
                    + "\t- faces.shape: {}\n".format(indfaces.shape)
                    + "\t- unique faces.shape: {}".format(facesu.shape)
                    + "\t- duplicate facess indices: {}\n".format(dupf))
        if lc[2]:
            nfu = facesuu.size
            nodnotf = [ii for ii in range(0, nnodes) if ii not in facesuu]
            fnotn = [ii for ii in facesuu if ii < 0 or ii >= nnodes]
            msg += ("  Non-bijective nodes indices vs faces:\n"
                    + "\t- nb. nodes: {}\n".format(nnodes)
                    + "\t- nb. unique nodes index in faces: {}\n".format(nfu)
                    + "\t- nodes not in faces: {}\n".format(nodnotf)
                    + "\t- faces ind not in nodes: {}\n".format(fnotn))
        raise Exception(msg)

    # Test for unused nodes
    facesu = np.unique(indfaces)
    c0 = np.all(facesu >= 0) and facesu.size == nnodes
    if not c0:
        indnot = [ii for ii in range(0, nnodes) if ii not in facesu]
        msg = ("Some nodes not used in mesh of ids {}:\n".format(ids)
               + "    - unused nodes indices: {}".format(indnot))
        warnings.warn(msg)

    # Convert to triangular mesh if necessary
    if mtype == 'quad':
        # Convert to tri mesh (solution for unstructured meshes)
        indface = np.empty((indfaces.shape[0]*2, 3), dtype=int)

        indface[::2, :] = indfaces[:, :3]
        indface[1::2, :-1] = indfaces[:, 2:]
        indface[1::2, -1] = indfaces[:, 0]
        indfaces = indface
        mtype = 'quadtri'
        ntri = 2
    else:
        ntri = 1

    # Check orientation
    x, y = nodes[indfaces, 0], nodes[indfaces, 1]
    orient = ((y[:, 1] - y[:, 0])*(x[:, 2] - x[:, 1])
              - (y[:, 2] - y[:, 1])*(x[:, 1] - x[:, 0]))

    indclock = orient > 0.
    if np.any(indclock):
        nclock, ntot = indclock.sum(), indfaces.shape[0]
        msg = ("Some triangles not counter-clockwise\n"
               + "  (necessary for matplotlib.tri.Triangulation)\n"
               + "    => {}/{} triangles reshaped".format(nclock, ntot))
        warnings.warn(msg)
        (indfaces[indclock, 1],
         indfaces[indclock, 2]) = indfaces[indclock, 2], indfaces[indclock, 1]
    return indfaces, mtype, ntri


# #############################################################################
#                   Rectangular meshes
# #############################################################################


def _rect_checkRZ(aa, name='R', shapeRZ=None):
    if aa.ndim == 1 and np.any(np.diff(aa) < 0.):
        msg = "{} must be increasing!".format(name)
        raise Exception(msg)
    elif aa.ndim == 2:
        lc = [np.all(np.diff(aa[0, :])) > 0.,
              np.all(np.diff(aa[:, 0])) > 0.]
        if np.sum(lc) != 1:
            msg = "{} must have exactly one dim increasing".format(name)
            raise Exception(msg)
        if lc[0]:
            aa = aa[0, :]
            if shapeRZ[1] is None:
                shapeRZ[1] = name
            if shapeRZ[1] != name:
                msg = ("Inconsistent shapeRZ[1]\n"
                       + "\t- expected: [{}, ...]\n".format(name)
                       + "\t- provided: {}".format(shapeRZ))
                raise Exception(msg)
        else:
            aa = aa[:, 0]
            if shapeRZ[0] is None:
                shapeRZ[0] = name
            assert shapeRZ[0] == name
    return aa, shapeRZ


def rect_checkformat(R, Z, datashape=None,
                     shapeRZ=None, ids=None):
    if R.ndim not in [1, 2] or Z.ndim not in [1, 2]:
        msg = ""
        raise Exception(msg)

    shapeu = np.unique(np.r_[R.shape, Z.shape])
    if shapeRZ is None:
        shapeRZ = [None, None]

    # Check R, Z
    R, shapeRZ = _rect_checkRZ(R, name='R', shapeRZ=shapeRZ)
    Z, shapeRZ = _rect_checkRZ(Z, name='Z', shapeRZ=shapeRZ)

    if datashape is not None:
        if None in shapeRZ:
            pass
        shapeRZ = tuple(shapeRZ)

        if shapeRZ == ('R', 'Z'):
            datashape_exp = (R.size, Z.size)
        elif shapeRZ == ('Z', 'R'):
            datashape_exp = (Z.size, R.size)
        else:
            msg = "Inconsistent data shape !"
            raise Exception(msg)
        if datashape != datashape_exp:
            msg = ("Inconsistent data shape\n"
                   + "\t- shapeRZ = {}\n".format(shapeRZ)
                   + "\t- datashape expected: {}\n".format(datashape_exp)
                   + "\t- datashape provided: {}\n".format(datashape))
            raise Exception(msg)

    if None not in shapeRZ:
        shapeRZ = tuple(shapeRZ)
        if shapeRZ not in [('R', 'Z'), ('Z', 'R')]:
            msg = ("Wrong value for shapeRZ:\n"
                   + "\t- expected: ('R', 'Z') or ('Z', 'R')\n"
                   + "\t- provided: {}".format(shapeRZ))
            raise Exception(msg)
    return R, Z, shapeRZ, 0



"""
That's where the default shortcuts for imas2tofu are defined
Shortcuts allow you (via the class MulTiIDSLoader) to define short str
    for some important data contained in IMAS ids
The are stored as a large dict, with:
    {ids: {
        shortcut1: long_version1,
        ...
        shortcutN: long_version1}
        }

There is a default copy of this file in the tfu install, but each user can
(re)define his/her own shortcuts using a local copy that will take precedence
at import time.

To customize your tofu install with user-specific parameters, run in terminal:

    tofu-custom

This will create a .tofu/ in your home (~) with a local copy that you can edit
To see the shortcuts available from your running ipython console do:

    > import tof as tf
    > tf.imas2tofu.MultiIDSLoader.get_shortcuts()

Available since tofu 1.4.3
"""

import numpy as np


# ############################################################################
#
#           General imas2tofu parameters
#
# ############################################################################


# public imas user (used for checking if can be saved)
_IMAS_USER_PUBLIC = 'imas_public'

# generic imas parameters dict
_IMAS_DIDD = {
    'shot': 0,
    'run': 0,
    'refshot': -1,
    'refrun': -1,
    'user': _IMAS_USER_PUBLIC,
    'tokamak': 'west',
    'version': '3',
}


# ############################################################################
#
#           shortcuts for imas2tofu interface (MultiIDSLoader class)
#
# ############################################################################

_dshort = {
    'wall': {
        'wallR': {'str': 'description_2d[0].limiter.unit[0].outline.r',
                  'units': 'm'},
        'wallZ': {'str': 'description_2d[0].limiter.unit[0].outline.z',
                  'units': 'm'}},

    'pulse_schedule': {
        'events_times': {'str': 'event[].time_stamp',
                         'units': 's'},
        'events_names': {'str': 'event[].identifier'}},

    'equilibrium': {
        't': {'str': 'time', 'units': 's'},
        'ip': {'str': 'time_slice[time].global_quantities.ip',
               'dim': 'current', 'quant': 'Ip', 'units': 'A'},
        'q0': {'str': 'time_slice[time].global_quantities.q_axis',
               'units': '-'},
        'qmin': {'str': 'time_slice[time].global_quantities.q_min.value',
                 'units': '-'},
        'q95': {'str': 'time_slice[time].global_quantities.q_95',
                'units': '-'},
        'volume': {'str': 'time_slice[time].global_quantities.volume',
                   'dim': 'volume', 'quant': 'pvol', 'units': 'm^3'},
        'psiaxis': {'str': 'time_slice[time].global_quantities.psi_axis',
                    'dim': 'B flux', 'quant': 'psi', 'units': 'Wb'},
        'psisep': {'str': 'time_slice[time].global_quantities.psi_boundary',
                   'dim': 'B flux', 'quant': 'psi', 'units': 'Wb'},
        'BT0': {'str': ('time_slice[time].global_quantities'
                        + '.magnetic_axis.b_field_tor'),
                'dim': 'B', 'quant': 'BT', 'units': 'T'},
        'axR': {'str': 'time_slice[time].global_quantities.magnetic_axis.r',
                'dim': 'distance', 'quant': 'R', 'units': 'm'},
        'axZ': {'str': 'time_slice[time].global_quantities.magnetic_axis.z',
                'dim': 'distance', 'quant': 'Z', 'units': 'm'},
        'x0R': {'str': 'time_slice[time].boundary.x_point[0].r', 'units': 'm'},
        'x0Z': {'str': 'time_slice[time].boundary.x_point[0].z', 'units': 'm'},
        'x1R': {'str': 'time_slice[time].boundary.x_point[1].r', 'units': 'm'},
        'x1Z': {'str': 'time_slice[time].boundary.x_point[1].z', 'units': 'm'},
        'strike0R': {'str': 'time_slice[time].boundary.strike_point[0].r',
                     'units': 'm'},
        'strike0Z': {'str': 'time_slice[time].boundary.strike_point[0].z',
                     'units': 'm'},
        'strike1R': {'str': 'time_slice[time].boundary.strike_point[1].r',
                     'units': 'm'},
        'strike1Z': {'str': 'time_slice[time].boundary.strike_point[1].z',
                     'units': 'm'},
        'sepR': {'str': 'time_slice[time].boundary_separatrix.outline.r',
                 'units': 'm'},
        'sepZ': {'str': 'time_slice[time].boundary_separatrix.outline.z',
                 'units': 'm'},

        '1drhotn': {'str': 'time_slice[time].profiles_1d.rho_tor_norm',
                    'dim': 'rho', 'quant': 'rhotn', 'units': '-'},
        '1dphi': {'str': 'time_slice[time].profiles_1d.phi',
                  'dim': 'B flux', 'quant': 'phi', 'units': 'Wb'},
        '1dpsi': {'str': 'time_slice[time].profiles_1d.psi',
                  'dim': 'B flux', 'quant': 'psi', 'units': 'Wb'},
        '1dq': {'str': 'time_slice[time].profiles_1d.q',
                'dim': 'safety factor', 'quant': 'q', 'units': '-'},
        '1dpe': {'str': 'time_slice[time].profiles_1d.pressure',
                 'dim': 'pressure', 'quant': 'pe', 'units': 'Pa'},
        '1djT': {'str': 'time_slice[time].profiles_1d.j_tor',
                 'dim': 'vol. current dens.', 'quant': 'jT',
                 'units': 'A.m^-2'},

        '2dphi': {'str': 'time_slice[time].ggd[0].phi[0].values',
                  'dim': 'B flux', 'quant': 'phi', 'units': 'Wb'},
        '2dpsi': {'str': 'time_slice[time].ggd[0].psi[0].values',
                  'dim': 'B flux', 'quant': 'psi', 'units': 'Wb'},
        '2djT': {'str': 'time_slice[time].ggd[0].j_tor[0].values',
                 'dim': 'vol. current dens.', 'quant': 'jT',
                 'units': 'A.m^-2'},
        '2dBR': {'str': 'time_slice[time].ggd[0].b_field_r[0].values',
                 'dim': 'B', 'quant': 'BR', 'units': 'T'},
        '2dBT': {'str': 'time_slice[time].ggd[0].b_field_tor[0].values',
                 'dim': 'B', 'quant': 'BT', 'units': 'T'},
        '2dBZ': {'str': 'time_slice[time].ggd[0].b_field_z[0].values',
                 'dim': 'B', 'quant': 'BZ', 'units': 'T'},
        '2dmeshNodes': {'str': ('grids_ggd[0].grid[0].space[0]'
                                + '.objects_per_dimension[0]'
                                + '.object[].geometry'),
                        'units': 'mixed'},
        '2dmeshFaces': {'str': ('grids_ggd[0].grid[0].space[0]'
                                + '.objects_per_dimension[2]'
                                + '.object[].nodes')},
        '2dmeshR': {'str': 'time_slice[0].profiles_2d[0].r', 'units': 'm'},
        '2dmeshZ': {'str': 'time_slice[0].profiles_2d[0].z', 'units': 'm'}},

    'core_profiles': {
        't': {'str': 'time', 'units': 's'},
        'ip': {'str': 'global_quantities.ip',
               'dim': 'current', 'quant': 'Ip', 'units': 'A'},
        'vloop': {'str': 'global_quantities.v_loop',
                  'dim': 'voltage', 'quant': 'Vloop', 'units': 'V'},

        '1dTe': {'str': 'profiles_1d[time].electrons.temperature',
                 'dim': 'temperature',  'quant': 'Te', 'units': 'eV'},
        '1dne': {'str': 'profiles_1d[time].electrons.density',
                 'dim': 'density', 'quant': 'ne', 'units': 'm^-3'},
        '1dzeff': {'str': 'profiles_1d[time].zeff',
                   'dim': 'charge', 'quant': 'zeff', 'units': '-'},
        '1dpsi': {'str': 'profiles_1d[time].grid.psi',
                  'dim': 'B flux', 'quant': 'psi', 'units': 'Wb'},
        '1drhotn': {'str': 'profiles_1d[time].grid.rho_tor_norm',
                    'dim': 'rho', 'quant': 'rhotn', 'units': '-'},
        '1drhopn': {'str': 'profiles_1d[time].grid.rho_pol_norm',
                    'dim': 'rho', 'quant': 'rhopn', 'units': '-'},
        '1dnW': {'str': 'profiles_1d[time].ion[identifier.label=W].density',
                 'dim': 'density', 'quant': 'nI', 'units': 'm^-3'}},

    'edge_profiles': {
        't': {'str': 'time', 'units': 's'}},

    'core_sources': {
        't': {'str': 'time', 'units': 's'},
        '1dpsi': {'str': ('source[identifier.name=lineradiation]'
                          + '.profiles_1d[time].grid.psi'),
                  'dim': 'B flux', 'quant': 'psi', 'units': 'Wb'},
        '1drhotn': {'str': ('source[identifier.name=lineradiation]'
                            + '.profiles_1d[time].grid.rho_tor_norm'),
                    'dim': 'rho', 'quant': 'rhotn', 'units': '-'},
        '1dbrem': {'str': ('source[identifier.name=bremsstrahlung]'
                           + '.profiles_1d[time].electrons.energy'),
                   'dim': 'vol.emis.', 'quant': 'brem.', 'units': 'W.m^-3'},
        '1dline': {'str': ('source[identifier.name=lineradiation]'
                           + '.profiles_1d[time].electrons.energy'),
                   'dim': 'vol. emis.', 'quant': 'lines', 'units': 'W.m^-3'}},

    'edge_sources': {
        't': {'str': 'time', 'units': 's'},
        '2dmeshNodes': {'str': ('grid_ggd[0].space[0].objects_per_dimension[0]'
                                + '.object[].geometry'),
                        'units': 'mixed'},
        '2dmeshFaces': {'str': ('grid_ggd[0].space[0].objects_per_dimension[2]'
                                + '.object[].nodes')},
        '2dradiation': {'str': 'source[13].ggd[0].electrons.energy[0].values',
                        'dim': 'vol. emis.', 'quant': 'vol.emis.',
                        'name': 'tot. vol. emis.', 'units': 'W.m^-3'}},

    'lh_antennas': {
        't': {'str': 'antenna[chan].power_launched.time', 'units': 's'},
        'power0': {'str': 'antenna[0].power_launched.data',
                   'dim': 'power', 'quant': 'lh power', 'units': 'W',
                   'pos': True},
        'power1': {'str': 'antenna[1].power_launched.data',
                   'dim': 'power', 'quant': 'lh power', 'units': 'W',
                   'pos': True},
        'power': {'str': 'antenna[chan].power_launched.data',
                  'dim': 'power', 'quant': 'lh power', 'units': 'W',
                  'pos': True},
        'R': {'str': 'antenna[chan].position.r.data',
              'dim': 'distance', 'quant': 'R', 'units': 'm'}},

    'ic_antennas': {
        't': {'str': 'antenna[chan].module[0].power_forward.time',
              'units': 's'},
        'power0mod_fwd': {'str': 'antenna[0].module[].power_forward.data',
                          'dim': 'power', 'quant': 'ic power', 'units': 'W'},
        'power0mod_reflect': {'str': ('antenna[0].module[]'
                                      + '.power_reflected.data'),
                              'dim': 'power', 'quant': 'ic power',
                              'units': 'W'},
        'power1mod_fwd': {'str': 'antenna[1].module[].power_forward.data',
                          'dim': 'power', 'quant': 'ic power', 'units': 'W'},
        'power1mod_reflect': {'str': ('antenna[1].module[]'
                                      + '.power_reflected.data'),
                              'dim': 'power', 'quant': 'ic power',
                              'units': 'W'},
        'power2mod_fwd': {'str': 'antenna[2].module[].power_forward.data',
                          'dim': 'power', 'quant': 'ic power', 'units': 'W'},
        'power2mod_reflect': {'str': ('antenna[2].module[]'
                                      + '.power_reflected.data'),
                              'dim': 'power', 'quant': 'ic power',
                              'units': 'W'}},

    'magnetics': {
        't': {'str': 'time', 'units': 's'},
        'ip': {'str': 'method[0].ip.data', 'units': 'A'},
        'diamagflux': {'str': 'method[0].diamagnetic_flux.data',
                       'units': 'Wb'},
        'bpol_B': {'str': 'bpol_probe[chan].field.data',
                   'dim': 'B', 'quant': 'Bpol', 'units': 'T'},
        'bpol_name': {'str': 'bpol_probe[chan].name'},
        'bpol_R': {'str': 'bpol_probe[chan].position.r',
                   'dim': 'distance', 'quant': 'R', 'units': 'm'},
        'bpol_Z': {'str': 'bpol_probe[chan].position.z',
                   'dim': 'distance', 'quant': 'Z', 'units': 'm'},
        'bpol_angpol': {'str': 'bpol_probe[chan].poloidal_angle',
                        'dim': 'angle', 'quant': 'angle_pol', 'units': 'rad'},
        'bpol_angtor': {'str': 'bpol_probe[chan].toroidal_angle',
                        'dim': 'angle', 'quant': 'angle_tor', 'units': 'rad'},
        'floop_flux': {'str': 'flux_loop[chan].flux.data',
                       'dim': 'B flux', 'quant': 'B flux', 'units': 'Wb'},
        'floop_name': {'str': 'flux_loop[chan].name'},
        'floop_R': {'str': 'flux_loop[chan].position.r',
                    'dim': 'distance', 'quant': 'R', 'units': 'm'},
        'floop_Z': {'str': 'flux_loop[chan].position.z',
                    'dim': 'distance', 'quant': 'Z', 'units': 'm'}},

    'barometry': {
        't': {'str': 'gauge[chan].pressure.time', 'units': 's'},
        'names': {'str': 'gauge[chan].name'},
        'p': {'str': 'gauge[chan].pressure.data',
              'dim': 'pressure', 'quant': 'p', 'units': 'Pa'}},

    'calorimetry': {
        't': {'str': 'group[chan].component[0].power.time', 'units': 's'},
        'names': {'str': 'group[chan].name'},
        'power': {'str': 'group[chan].component[0].power.data',
                  'dim': 'power', 'quant': 'extracted power',
                  'units': 'W'}},

    'neutron_diagnostic': {
        't': {'str': 'time', 'units': 's'},
        'flux_total': {'str': 'synthetic_signals.total_neutron_flux',
                       'dim': 'particle flux', 'quant': 'particle flux',
                       'units': 's^-1'}},

    'ece': {
        't': {'str': 'time',
              'quant': 't', 'units': 's'},
        'freq': {'str': 'channel[chan].frequency.data',
                 'dim': 'freq', 'quant': 'freq', 'units': 'Hz'},
        'Te': {'str': 'channel[chan].t_e.data',
               'dim': 'temperature', 'quant': 'Te', 'units': 'eV'},
        'R': {'str': 'channel[chan].position.r',
              'dim': 'distance', 'quant': 'R', 'units': 'm'},
        'rhotn': {'str': 'channel[chan].position.rho_tor_norm',
                  'dim': 'rho', 'quant': 'rhotn', 'units': '-'},
        'theta': {'str': 'channel[chan].position.theta',
                  'dim': 'angle', 'quant': 'theta', 'units': 'rad'},
        'tau1keV': {'str': 'channel[chan].optical_depth.data',
                    'dim': 'optical_depth', 'quant': 'tau', 'units': '-'},
        'validity_timed': {'str': 'channel[chan].t_e.validity_timed'},
        'names': {'str': 'channel[chan].name'},
        'Te0': {'str': 't_e_central.data',
                'dim': 'temperature', 'quant': 'Te', 'units': 'eV'}},

    'reflectometer_profile': {
        't': {'str': 'time', 'units': 's'},
        'ne': {'str': 'channel[chan].n_e.data',
               'dim': 'density', 'quant': 'ne', 'units': 'm^-3'},
        'R': {'str': 'channel[chan].position.r',
              'dim': 'distance', 'quant': 'R', 'units': 'm'},
        'Z': {'str': 'channel[chan].position.z',
              'dim': 'distance', 'quant': 'Z', 'units': 'm'},
        'phi': {'str': 'channel[chan].position.phi',
                'dim': 'angle', 'quant': 'phi', 'units': 'rad'},
        'names': {'str': 'channel[chan].name'},
        'mode': {'str': 'mode'},
        'sweep': {'str': 'sweep_time'}},

    'interferometer': {
        't': {'str': 'time',
              'quant': 't', 'units': 's'},
        'names': {'str': 'channel[chan].name'},
        'ne_integ': {'str': 'channel[chan].n_e_line.data',
                     'dim': 'ne_integ', 'quant': 'ne_integ',
                     'units': 'm^-2', 'Brightness': True}},

    'polarimeter': {
        't': {'str': 'time',
              'quant': 't', 'units': 's'},
        'lamb': {'str': 'channel[chan].wavelength',
                 'dim': 'distance', 'quant': 'wavelength',
                 'units': 'm'},
        'fangle': {'str': 'channel[chan].faraday_angle.data',
                   'dim': 'angle', 'quant': 'faraday angle',
                   'units': 'rad', 'Brightness': True},
        'names': {'str': 'channel[chan].name'}},

    'bolometer': {
        't': {'str': 'channel[chan].power.time',
              'quant': 't', 'units': 's'},
        'power': {'str': 'channel[chan].power.data',
                  'dim': 'power', 'quant': 'power radiative',
                  'units': 'W', 'Brightness': False},
        'etendue': {'str': 'channel[chan].etendue',
                    'dim': 'etendue', 'quant': 'etendue',
                    'units': 'm^2.sr'},
        'names': {'str': 'channel[chan].name'},
        'tpower': {'str': 'time', 'quant': 't', 'units': 's'},
        'prad': {'str': 'power_radiated_total',
                 'dim': 'power', 'quant': 'power radiative',
                 'units': 'W'},
        'pradbulk': {'str': 'power_radiated_inside_lcfs',
                     'dim': 'power', 'quant': 'power radiative',
                     'units': 'W'}},

    'soft_x_rays': {
        't': {'str': 'time',
              'quant': 't', 'units': 's'},
        'power': {'str': 'channel[chan].power.data',
                  'dim': 'power', 'quant': 'power radiative',
                  'units': 'W', 'Brightness': False},
        'brightness': {'str': 'channel[chan].brightness.data',
                       'dim': 'brightness', 'quant': 'brightness',
                       'units': 'W.m^-2.sr^-1', 'Brightness': True},
        'names': {'str': 'channel[chan].name'},
        'etendue': {'str': 'channel[chan].etendue',
                    'dim': 'etendue', 'quant': 'etendue',
                    'units': 'm^2.sr'}},

    'spectrometer_visible': {
        't': {'str': ('channel[chan].grating_spectrometer'
                      + '.radiance_spectral.time'),
              'quant': 't', 'units': 's'},
        'spectra': {'str': ('channel[chan].grating_spectrometer'
                            + '.radiance_spectral.data'),
                    'dim': 'radiance_spectral',
                    'quant': 'radiance_spectral',
                    'units': '(photons).m^-2.s^-1.sr^-1.m^-1',
                    'Brightness': True},
        'names': {'str': 'channel[chan].name'},
        'lamb': {'str': 'channel[chan].grating_spectrometer.wavelengths',
                 'dim': 'wavelength', 'quant': 'wavelength', 'units': 'm'}},

    'bremsstrahlung_visible': {
        't': {'str': 'time',
              'quant': 't', 'units': 's'},
        'radiance': {'str': 'channel[chan].radiance_spectral.data',
                     'dim': 'radiance_spectral',
                     'quant': 'radiance_spectral',
                     'units': '(photons).m^-2.s^-1.sr^-1.m^-1',
                     'Brightness': True},
        'names': {'str': 'channel[chan].name'},
        'lamb_up': {'str': 'channel[chan].filter.wavelength_upper',
                    'units': 'm'},
        'lamb_lo': {'str': 'channel[chan].filter.wavelength_lower',
                    'units': 'm'}}
    }


# ############################################################################
#
#           default data for each ids (not used yet)
#
# ############################################################################


_didsdiag = {
    'lh_antennas': {'datacls': 'DataCam1D',
                    'geomcls': False,
                    'sig': {'data': 'power',
                            't': 't'}},
    'ic_antennas': {'datacls': 'DataCam1D',
                    'geomcls': False,
                    'sig': {'data': 'power',
                            't': 't'}},
    'magnetics': {'datacls': 'DataCam1D',
                  'geomcls': False,
                  'sig': {'data': 'bpol_B',
                          't': 't'}},
    'barometry': {'datacls': 'DataCam1D',
                  'geomcls': False,
                  'sig': {'data': 'p',
                          't': 't'}},
    'calorimetry': {'datacls': 'DataCam1D',
                    'geomcls': False,
                    'sig': {'data': 'power',
                            't': 't'}},
    'ece': {'datacls': 'DataCam1D',
            'geomcls': False,
            'sig': {'t': 't',
                    'X': 'rhotn_sign',
                    'data': 'Te'},
            'stack': True},
    'neutron_diagnostic': {'datacls': 'DataCam1D',
                           'geomcls': False,
                           'sig': {'t': 't',
                                   'data': 'flux_total'}},
    'reflectometer_profile': {'datacls': 'DataCam1D',
                              'geomcls': False,
                              'sig': {'t': 't',
                                      'X': 'R',
                                      'data': 'ne'}},
    'interferometer': {'datacls': 'DataCam1D',
                       'geomcls': 'CamLOS1D',
                       'sig': {'t': 't',
                               'data': 'ne_integ'},
                       'synth': {'dsynth': {
                           'quant': 'core_profiles.1dne',
                           'ref1d': 'core_profiles.1drhotn',
                           'ref2d': 'equilibrium.2drhotn'},
                               'dsig': {'core_profiles': ['t'],
                                        'equilibrium': ['t']},
                               'Brightness': True},
                       'stack': True},
    'polarimeter': {'datacls': 'DataCam1D',
                    'geomcls': 'CamLOS1D',
                    'sig': {'t': 't',
                            'data': 'fangle'},
                    'synth': {'dsynth': {
                        'fargs': ['core_profiles.1dne',
                                  'equilibrium.2dBR',
                                  'equilibrium.2dBT',
                                  'equilibrium.2dBZ',
                                  'core_profiles.1drhotn',
                                  'equilibrium.2drhotn']},
                              'dsig': {'core_profiles': ['t'],
                                       'equilibrium': ['t']},
                              'Brightness': True},
                    'stack': True},
    'bolometer': {'datacls': 'DataCam1D',
                  'geomcls': 'CamLOS1D',
                  'sig': {'t': 't',
                          'data': 'power'},
                  'synth': {'dsynth': {
                      'quant': 'core_sources.1dprad',
                      'ref1d': 'core_sources.1drhotn',
                      'ref2d': 'equilibrium.2drhotn'},
                            'dsig': {'core_sources': ['t'],
                                     'equilibrium': ['t']},
                            'Brightness': False},
                  'stack': True},
    'soft_x_rays': {'datacls': 'DataCam1D',
                    'geomcls': 'CamLOS1D',
                    'sig': {'t': 't',
                            'data': 'power'},
                    'stack': True},
    'spectrometer_visible': {'datacls': 'DataCam1DSpectral',
                             'geomcls': 'CamLOS1D',
                             'sig': {'data': 'spectra',
                                     't': 't',
                                     'lamb': 'lamb'}},
    'bremsstrahlung_visible': {
        'datacls': 'DataCam1D',
        'geomcls': 'CamLOS1D',
        'sig': {'t': 't',
                'data': 'radiance'},
        'synth': {
            'dsynth': {
                'quant': ['core_profiles.1dTe',
                          'core_profiles.1dne',
                          'core_profiles.1dzeff'],
                'ref1d': 'core_profiles.1drhotn',
                'ref2d': 'equilibrium.2drhotn'},
            'dsig': {'core_profiles': ['t'],
                     'equilibrium': ['t']},
            'Brightness': True},
        'stack': True}
    }


# ############################################################################
#
#           Complete dshort and didsdiag
#
# ############################################################################


_lidsdiag = sorted([kk for kk, vv in _didsdiag.items() if 'sig' in vv.keys()])
_lidslos = list(_lidsdiag)
for ids_ in _lidsdiag:
    if _didsdiag[ids_]['geomcls'] not in ['CamLOS1D']:
        _lidslos.remove(ids_)

for ids in _lidslos:
    dlos = {}
    strlos = 'line_of_sight'
    if ids == 'reflectometer_profile':
        strlos += '_detection'
    dlos['los_pt1R'] = {
        'str': 'channel[chan].{}.first_point.r'.format(strlos),
        'units': 'm'}
    dlos['los_pt1Z'] = {
        'str': 'channel[chan].{}.first_point.z'.format(strlos),
        'units': 'm'}
    dlos['los_pt1Phi'] = {
        'str': 'channel[chan].{}.first_point.phi'.format(strlos),
        'units': 'rad'}
    dlos['los_pt2R'] = {
        'str': 'channel[chan].{}.second_point.r'.format(strlos),
        'units': 'm'}
    dlos['los_pt2Z'] = {
        'str': 'channel[chan].{}.second_point.z'.format(strlos),
        'units': 'm'}
    dlos['los_pt2Phi'] = {
        'str': 'channel[chan].{}.second_point.phi'.format(strlos),
        'units': 'rad'}
    _dshort[ids].update(dlos)


_lidssynth = sorted([kk for kk, vv in _didsdiag.items()
                     if 'synth' in vv.keys()])
for ids_ in _lidssynth:
    for kk, vv in _didsdiag[ids_]['synth']['dsynth'].items():
        if type(vv) is str:
            vv = [vv]
        for ii in range(0, len(vv)):
            v0, v1 = vv[ii].split('.')
            if v0 not in _didsdiag[ids_]['synth']['dsig'].keys():
                _didsdiag[ids_]['synth']['dsig'][v0] = [v1]
            elif v1 not in _didsdiag[ids_]['synth']['dsig'][v0]:
                _didsdiag[ids_]['synth']['dsig'][v0].append(v1)
        _didsdiag[ids_]['synth']['dsynth'][kk] = vv


# ############################################################################
#
#           Dict for computing signals from loaded signals
#
# ############################################################################


# -------------
# Functions

def _events(names, t):
    ustr = 'U{}'.format(np.nanmax(np.char.str_len(np.char.strip(names))))
    return np.array([(nn, tt)
                     for nn, tt in zip(*[np.char.strip(names), t])],
                    dtype=[('name', ustr), ('t', np.float)])


def _RZ2array(ptsR, ptsZ):
    out = np.array([ptsR, ptsZ]).T
    if out.ndim == 1:
        out = out[None, :]
    return out


def _losptsRZP(*pt12RZP):
    return np.swapaxes([pt12RZP[:3], pt12RZP[3:]], 0, 1).T


def _add(a0, a1):
    return np.abs(a0 + a1)


def _eqB(BT, BR, BZ):
    return np.sqrt(BT**2 + BR**2 + BZ**2)


def _icmod(al, ar, axis=0):
    return np.sum(al - ar, axis=axis)


def _icmodadd(al0, ar0, al1, ar1, al2, ar2, axis=0):
    return (np.sum(al0 - ar0, axis=axis)
            + np.sum(al1 - ar1, axis=axis)
            + np.sum(al2 - ar2, axis=axis))


def _rhopn1d(psi):
    return np.sqrt((psi - psi[:, 0:1]) / (psi[:, -1] - psi[:, 0])[:, None])


def _rhopn2d(psi, psi0, psisep):
    return np.sqrt(
        (psi - psi0[:, None]) / (psisep[:, None] - psi0[:, None]))


def _rhotn2d(phi):
    return np.sqrt(np.abs(phi) / np.nanmax(np.abs(phi), axis=1)[:, None])


def _eqSep(sepR, sepZ, npts=100):
    nt = len(sepR)
    assert len(sepZ) == nt
    sep = np.full((nt, npts, 2), np.nan)
    pts = np.linspace(0, 100, npts)
    for ii in range(0, nt):
        ptsii = np.linspace(0, 100, sepR[ii].size)
        sep[ii, :, 0] = np.interp(pts, ptsii, sepR[ii])
        sep[ii, :, 1] = np.interp(pts, ptsii, sepZ[ii])
    return sep


def _eqtheta(axR, axZ, nodes, cocos=11):
    theta = np.arctan2(nodes[:, 0][None, :] - axZ[:, None],
                       nodes[:, 1][None, :] - axR[:, None])
    if cocos == 1:
        theta = -theta
    return theta


def _rhosign(rho, theta):
    if isinstance(theta, np.ndarray):
        rhotns = np.array(rho)
        ind = ~np.isnan(theta)
        ind[ind] &= np.cos(theta[ind]) < 0.
        rhotns[ind] = -rho[ind]
    else:
        rhotns = [None for ii in range(len(theta))]
        for ii in range(len(theta)):
            rhotns[ii] = _rhosign(rho[ii], theta[ii])
    return rhotns


def _lamb(lamb_up, lamb_lo):
    return 0.5*(lamb_up + lamb_lo)


# ----------
# dict


_dcomp = {
          'pulse_schedule':
          {'events': {'lstr': ['events_names', 'events_times'],
                      'func': _events}},

          'wall':
          {'wall': {'lstr': ['wallR', 'wallZ'],
                    'func': _RZ2array}},

          'equilibrium':
          {'ax': {'lstr': ['axR', 'axZ'], 'func': _RZ2array, 'units': 'm'},
           'sep': {'lstr': ['sepR', 'sepZ'],
                   'func': _eqSep, 'kargs': {'npts': 100}, 'units': 'm'},
           '2dB': {'lstr': ['2dBT', '2dBR', '2dBZ'], 'func': _eqB,
                   'dim': 'B', 'quant': 'B', 'units': 'T'},
           '1drhopn': {'lstr': ['1dpsi', 'psiaxis', 'psisep'],
                       'func': _rhopn2d,
                       'dim': 'rho', 'quant': 'rhopn', 'units': '-'},
           '2drhopn': {'lstr': ['2dpsi', 'psiaxis', 'psisep'],
                       'func': _rhopn2d,
                       'dim': 'rho', 'quant': 'rhopn', 'units': '-'},
           '2drhotn': {'lstr': ['2dphi'], 'func': _rhotn2d,
                       'dim': 'rho', 'quant': 'rhotn', 'units': '-'},
           'x0': {'lstr': ['x0R', 'x0Z'], 'func': _RZ2array, 'units': 'm'},
           'x1': {'lstr': ['x1R', 'x1Z'], 'func': _RZ2array, 'units': 'm'},
           'strike0': {'lstr': ['strike0R', 'strike0Z'], 'func': _RZ2array,
                       'units': 'm'},
           'strike1': {'lstr': ['strike1R', 'strike1Z'], 'func': _RZ2array,
                       'units': 'm'},
           '2dtheta': {'lstr': ['axR', 'axZ', '2dmeshNodes'],
                       'func': _eqtheta, 'kargs': {'cocos': 11},
                       'units': 'rad'}},

          'core_profiles':
          {'1drhopn': {'lstr': ['1dpsi'], 'func': _rhopn1d,
                       'dim': 'rho', 'quant': 'rhopn', 'units': '-'}},

          'core_sources':
          {'1drhopn': {'lstr': ['1dpsi'], 'func': _rhopn1d,
                       'dim': 'rho', 'quant': 'rhopn', 'units': '-'},
           '1dprad': {'lstr': ['1dbrem', '1dline'], 'func': _add,
                      'dim': 'vol. emis.', 'quant': 'prad',
                      'units': 'W.m^-3'}},

          'magnetics':
          {'bpol_pos': {'lstr': ['bpol_R', 'bpol_Z'], 'func': _RZ2array},
           'floop_pos': {'lstr': ['floop_R', 'floop_Z'], 'func': _RZ2array}},

          'ic_antennas': {
              'power0': {'lstr': ['power0mod_fwd', 'power0mod_reflect'],
                         'func': _icmod, 'kargs': {'axis': 0},
                         'pos': True, 'units': 'W'},
              'power1': {'lstr': ['power1mod_fwd', 'power1mod_reflect'],
                         'func': _icmod, 'kargs': {'axis': 0},
                         'pos': True, 'units': 'W'},
              'power2': {'lstr': ['power2mod_fwd', 'power2mod_reflect'],
                         'func': _icmod, 'kargs': {'axis': 0},
                         'pos': True, 'units': 'W'},
              'power': {'lstr': ['power0mod_fwd', 'power0mod_reflect',
                                 'power1mod_fwd', 'power1mod_reflect',
                                 'power2mod_fwd', 'power2mod_reflect'],
                        'func': _icmodadd, 'kargs': {'axis': 0},
                        'pos': True, 'units': 'W'}},

          'ece':
          {'rhotn_sign': {'lstr': ['rhotn', 'theta'], 'func': _rhosign,
                          'units': '-'}},

          'bremsstrahlung_visible':
          {'lamb': {'lstr': ['lamb_up', 'lamb_lo'], 'func': _lamb,
                    'dim': 'distance',
                    'quantity': 'wavelength',
                    'units': 'm'}}
          }

# Complete los
_lstr = ['los_pt1R', 'los_pt1Z', 'los_pt1Phi',
         'los_pt2R', 'los_pt2Z', 'los_pt2Phi']
for ids in _lidslos:
    _dcomp[ids] = _dcomp.get(ids, {})
    _dcomp[ids]['los_ptsRZPhi'] = {'lstr': _lstr, 'func': _losptsRZP}


# Uniformize
_lids = set(_dshort.keys()).union(_dcomp.keys())
for ids in _lids:
    _dshort[ids] = _dshort.get(ids, {})
    _dcomp[ids] = _dcomp.get(ids, {})


# ############################################################################
#
#           Dict of signals that, by default are not loaded
#         (because replaced by more complete computed signals)
#                         (avoids redundancy)
#
# ############################################################################


# All except (for when sig not specified in get_data())
_dall_except = {}
for ids in _lidslos:
    _dall_except[ids] = _lstr
_dall_except['equilibrium'] = ['axR', 'axZ', 'sepR', 'sepZ',
                               '2dBT', '2dBR', '2dBZ',
                               'x0R', 'x0Z', 'x1R', 'x1Z',
                               'strike0R', 'strike0Z', 'strike1R', 'strike1Z']
_dall_except['magnetics'] = ['bpol_R', 'bpol_Z', 'floop_R', 'floop_Z']
_dall_except['ic_antennas'] = ['power0mod_launched', 'power0mod_reflected',
                               'power1mod_launched', 'power1mod_reflected',
                               'power2mod_launched', 'power2mod_reflected']


# ############################################################################
#
#           Dict of default pre-defined sets of signals
#
# ############################################################################


_dpreset = {
            'overview':
            {'wall': None,
             'pulse_schedule': None,
             'equilibrium': None},

            'plasma2d':
            {'wall': ['domainR', 'domainZ'],
             'equilibrium': ['t', 'ax', 'sep'],
             'core_profiles': ['t', '1dTe', '1dne', '1dzeff',
                               '1drhotn', '1dphi'],
             'core_sources': ['t', '1dprad'],
             'edge_profiles': ['t'],
             'edge_sources': ['t']},

            'ece':
            {'wall': ['domainR', 'domainZ'],
             'ece': None,
             'core_profiles': ['t', 'Te', 'ne']}
           }


# ############################################################################
#
#       List of ids considered as basis and automatically loaded by default
#
# ############################################################################


_IDS_BASE = ['wall', 'pulse_schedule']


# ############################################################################
#
#       Default parameters for exporting to tofu objects (_comp_toobjects.py)
#
# ############################################################################

_INDEVENT = 0
_DTLIM = {'west': ['IGNITRON', 'PLUSDIP']}



"""
Getting started: 5 minutes tutorial for `tofu`
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

This is a tutorial that aims to get a new user a little familiar with tofu's
 structure.
"""

# The following imports matplotlib, preferably using a
# backend that allows the plots to be interactive (Qt5Agg).
import matplotlib
try:
    matplotlib.use('Qt5Agg')
except ImportError:
    matplotlib.use(matplotlib.rcParams['backend'])

###############################################################################
# We start by loading `tofu`. You might see some warnings at this stage since
# optional modules for `tofu` could
# be missing on the machine you are working on. This can be ignored safely.

import numpy as np
import tofu as tf

###############################################################################
# We can now create our first configuration.
# In `tofu` speak, a configuration is the geometry
# of the device and its structures.
# `tofu` provides pre-defined ones for your to try,
# so we're going to do just that:

configB2 = tf.geom.utils.create_config("B2")

###############################################################################
# The configuration can easily be visualized using the `.plot()` method:

configB2.plot()

###############################################################################
# Since `tofu` is all about tomography,
# let's create a 1D camera and plot its output.

cam1d = tf.geom.utils.create_CamLOS1D(
    config=configB2,
    P=[3.4, 0, 0],
    N12=100,
    F=0.1,
    D12=0.1,
    angs=[np.pi, 0, 0],
    Name="",
    Exp="",
    Diag="",
)
# interactive plot
cam1d.plot_touch()

###############################################################################
# The principle is similar for 2D cameras.

cam2d = tf.geom.utils.create_CamLOS2D(
    config=configB2,
    P=[3.4, 0, 0],
    N12=100,
    F=0.1,
    D12=0.1,
    angs=[np.pi, 0, 0],
    Name="",
    Exp="",
    Diag="",
)
cam2d.plot_touch()

###############################################################################
# What comes next is up to you!
# You could now play with the function parameters (change the cameras
# direction, refinement, aperture),
# with the plots (many are interactive) or create you own tomographic
# configuration.





import numpy as np
from scipy.interpolate import BSpline
import matplotlib.pyplot as plt
from matplotlib.colors import ListedColormap
import matplotlib.gridspec as gridspec
from matplotlib.axes._axes import Axes
from mpl_toolkits.mplot3d import Axes3D

# tofu
from tofu.version import __version__
from . import _def as _def

_GITHUB = 'https://github.com/ToFuProject/tofu/issues'
_WINTIT = 'tofu-%s        report issues / requests at %s'%(__version__, _GITHUB)

_QUIVERCOLOR = plt.cm.viridis(np.linspace(0, 1, 3))
_QUIVERCOLOR = np.array([[1., 0., 0., 1.],
                         [0., 1., 0., 1.],
                         [0., 0., 1., 1.]])
_QUIVERCOLOR = ListedColormap(_QUIVERCOLOR)


# Generic
def _check_projdax_mpl(dax=None, proj=None, fs=None, wintit=None):

    # ----------------------
    # Check proj
    if proj is None:
        proj = 'all'
    assert isinstance(proj, str)
    proj = proj.lower()
    lproj = ['cross', 'hor', '3d']
    assert proj in lproj + ['all']
    if proj == 'all':
        proj = ['cross', 'hor']
    else:
        proj = [proj]

    # ----------------------
    # Check dax
    lc = [dax is None, issubclass(dax.__class__, Axes),
          isinstance(dax, dict), isinstance(dax, list)]
    assert any(lc)
    if lc[0]:
        dax = dict.fromkeys(proj)
    elif lc[1]:
        assert len(proj) == 1
        dax = {proj[0]: dax}
    elif lc[2]:
        lcax = [(kk in proj
                 and (ax is None or issubclass(ax.__class__, Axes)))
                for kk, ax in dax.items()]
        if not all(lcax):
            msg = "Wrong key or axes in dax:\n"
            msg += "    - proj = %s"%str(proj)
            msg += "    - dax = %s"%str(dax)
            raise Exception(msg)
    else:
        assert len(dax) == 2
        assert all([ax is None or issubclass(ax.__class__, Axes)
                    for ax in dax])
        dax = {'cross': dax[0], 'hor': dax[1]}

    # Populate with default axes if necessary
    if 'cross' in proj and 'hor' in proj:
        if 'cross' in proj and 'hor' in proj:
            if dax['cross'] is None:
                assert dax['hor'] is None
                lax = _def.Plot_LOSProj_DefAxes('all', fs=fs,                                                                 wintit=wintit)
                dax['cross'], dax['hor'] = lax
        elif 'cross' in proj and dax['cross'] is None:
            dax['cross'] = _def.Plot_LOSProj_DefAxes('cross', fs=fs,
                                                     wintit=wintit)
        elif 'hor' in proj and dax['hor'] is None:
            dax['hor'] = _def.Plot_LOSProj_DefAxes('hor', fs=fs,
                                                     wintit=wintit)
        elif '3d' in proj  and dax['3d'] is None:
            dax['3d'] = _def.Plot_3D_plt_Tor_DefAxes(fs=fs,
                                                    wintit=wintit)
    for kk in lproj:
        dax[kk] = dax.get(kk, None)
    return dax



# #################################################################
# #################################################################
#                   Generic geometry plot
# #################################################################
# #################################################################

def CrystalBragg_plot(cryst, lax=None, proj=None, res=None, element=None,
                      color=None, dP=None,
                      det_cent=None, det_nout=None, det_ei=None, det_ej=None,
                      dI=None, dBs=None, dBv=None,
                      dVect=None, dIHor=None, dBsHor=None, dBvHor=None,
                      dleg=None, indices=False,
                      draw=True, fs=None, wintit=None, tit=None, Test=True):

    # ---------------------
    # Check / format inputs

    if Test:
        msg = "Arg proj must be in ['cross','hor','all','3d'] !"
        assert type(draw) is bool, "Arg draw must be a bool !"
        assert cryst.__class__.__name__ == 'CrystalBragg'
    if wintit is None:
        wintit = _WINTIT
    if dleg is None:
         dleg = _def.TorLegd

    # ---------------------
    # call plotting functions

    kwa = dict(fs=fs, wintit=wintit, Test=Test)
    if proj == '3d':
        # Temporary matplotlib issue
        dleg = None
    else:
        dax = _CrystalBragg_plot_crosshor(cryst, proj=proj, res=res, dax=lax,
                                          element=element, color=color,
                                          det_cent=det_cent, det_nout=det_nout,
                                          det_ei=det_ei, det_ej=det_ej)

    # recompute the ax.dataLim
    ax0 = None
    for kk, vv in dax.items():
        if vv is None:
            continue
        dax[kk].relim()
        dax[kk].autoscale_view()
        if dleg is not None:
            dax[kk].legend(**dleg)
        ax0 = vv

    # set title
    if tit != False:
        ax0.figure.suptitle(tit)
    if draw:
        ax0.figure.canvas.draw()
    return dax


def _CrystalBragg_plot_crosshor(cryst, proj=None, dax=None,
                                element=None, res=None,
                                det_cent=None, det_nout=None,
                                det_ei=None, det_ej=None,
                                Pdict=_def.TorPd, Idict=_def.TorId, Bsdict=_def.TorBsd,
                                Bvdict=_def.TorBvd, Vdict=_def.TorVind,
                                color=None, ms=None, quiver_cmap=None,
                                LegDict=_def.TorLegd, indices=False,
                                draw=True, fs=None, wintit=None, Test=True):
    if Test:
        assert type(Pdict) is dict, 'Arg Pdict should be a dictionary !'
        assert type(Idict) is dict, "Arg Idict should be a dictionary !"
        assert type(Bsdict) is dict, "Arg Bsdict should be a dictionary !"
        assert type(Bvdict) is dict, "Arg Bvdict should be a dictionary !"
        assert type(Vdict) is dict, "Arg Vdict should be a dictionary !"
        assert type(LegDict) is dict or LegDict is None, 'Arg LegDict should be a dictionary !'

    # ---------------------
    # Check / format inputs

    if element is None:
        element = 'oscvr'
    element = element.lower()
    if 'v' in element and quiver_cmap is None:
        quiver_cmap = _QUIVERCOLOR
    if color is None:
        if cryst._dmisc.get('color') is not None:
            color = cryst._dmisc['color']
        else:
            color = 'k'
    if ms is None:
        ms = 6

    # ---------------------
    # Prepare axe and data

    dax = _check_projdax_mpl(dax=dax, proj=proj, fs=fs, wintit=wintit)

    if 's' in element or 'v' in element:
        summ = cryst._dgeom['summit']
    if 'c' in element:
        cent = cryst._dgeom['center']
    if 'r' in element:
        ang = np.linspace(0, 2.*np.pi, 200)
        rr = 0.5*cryst._dgeom['rcurve']
        row = cryst._dgeom['center'] + rr*cryst._dgeom['nout']
        row = (row[:, None]
               + rr*(np.cos(ang)[None, :]*cryst._dgeom['nout'][:, None]
                     + np.sin(ang)[None, :]*cryst._dgeom['e1'][:, None]))

    # ---------------------
    # plot

    if 'o' in element:
        cont = cryst.sample_outline_plot(res=res)
        if dax['cross'] is not None:
            dax['cross'].plot(np.hypot(cont[0,:], cont[1,:]), cont[2,:],
                              ls='-', c=color, marker='None',
                              label=cryst.Id.NameLTX+' contour')
        if dax['hor'] is not None:
            dax['hor'].plot(cont[0,:], cont[1,:],
                            ls='-', c=color, marker='None',
                            label=cryst.Id.NameLTX+' contour')
    if 's' in element:
        if dax['cross'] is not None:
            dax['cross'].plot(np.hypot(summ[0], summ[1]), summ[2],
                              marker='^', ms=ms, c=color,
                              label=cryst.Id.NameLTX+" summit")
        if dax['hor'] is not None:
            dax['hor'].plot(summ[0], summ[1],
                            marker='^', ms=ms, c=color,
                            label=cryst.Id.NameLTX+" summit")
    if 'c' in element:
        if dax['cross'] is not None:
            dax['cross'].plot(np.hypot(cent[0], cent[1]), cent[2],
                              marker='o', ms=ms, c=color,
                              label=cryst.Id.NameLTX+" center")
        if dax['hor'] is not None:
            dax['hor'].plot(cent[0], cent[1],
                            marker='o', ms=ms, c=color,
                            label=cryst.Id.NameLTX+" center")
    if 'r' in element:
        if dax['cross'] is not None:
            dax['cross'].plot(np.hypot(row[0,:], row[1,:]), row[2,:],
                              ls='--', color=color, marker='None',
                              label=cryst.Id.NameLTX+' rowland')
        if dax['hor'] is not None:
            dax['hor'].plot(row[0,:], row[1,:],
                            ls='--', color=color, marker='None',
                            label=cryst.Id.NameLTX+' rowland')
    if 'v' in element:
        nin = cryst._dgeom['nin']
        e1, e2 = cryst._dgeom['e1'], cryst._dgeom['e2']
        p0 = np.repeat(summ[:,None], 3, axis=1)
        v = np.concatenate((nin[:, None], e1[:, None], e2[:, None]), axis=1)
        if dax['cross'] is not None:
            pr = np.hypot(p0[0, :], p0[1, :])
            vr = np.hypot(p0[0, :]+v[0, :], p0[1, :]+v[1, :]) - pr
            dax['cross'].quiver(pr, p0[2, :],
                                vr, v[2, :],
                                np.r_[0., 0.5, 1.], cmap=quiver_cmap,
                                angles='xy', scale_units='xy',
                                label=cryst.Id.NameLTX+" unit vect", **Vdict)
        if dax['hor'] is not None:
            dax['hor'].quiver(p0[0, :], p0[1, :],
                              v[0, :], v[1, :],
                              np.r_[0., 0.5, 1.], cmap=quiver_cmap,
                              angles='xy', scale_units='xy',
                              label=cryst.Id.NameLTX+" unit vect", **Vdict)

    # Detector
    sc = None
    if det_cent is not None:
        if dax['cross'] is not None:
            dax['cross'].plot(np.hypot(det_cent[0], det_cent[1]), det_cent[2],
                              marker='x', ms=ms, c=color, label="det_cent")
        if dax['hor'] is not None:
            dax['hor'].plot(det_cent[0], det_cent[1],
                            marker='x', ms=ms, c=color, label="det_cent")

    if det_nout is not None:
        assert det_ei is not None and det_ej is not None
        p0 = np.repeat(det_cent[:, None], 3, axis=1)
        v = np.concatenate((det_nout[:, None], det_ei[:, None],
                            det_ej[:, None]), axis=1)
        if dax['cross'] is not None:
            pr = np.hypot(p0[0, :], p0[1, :])
            vr = np.hypot(p0[0, :]+v[0, :], p0[1, :]+v[1, :]) - pr
            dax['cross'].quiver(pr, p0[2, :],
                                vr, v[2, :],
                                np.r_[0., 0.5, 1.], cmap=quiver_cmap,
                                angles='xy', scale_units='xy',
                                label="det unit vect", **Vdict)
        if dax['hor'] is not None:
            dax['hor'].quiver(p0[0, :], p0[1, :],
                              v[0, :], v[1, :],
                              np.r_[0., 0.5, 1.], cmap=quiver_cmap,
                              angles='xy', scale_units='xy',
                              label="det unit vect", **Vdict)

    return dax


# #################################################################
# #################################################################
#                   Rocking curve plot
# #################################################################
# #################################################################

def CrystalBragg_plot_rockingcurve(Rmax=None, sigma=None,
                                   bragg=None, delta_bragg=None, npts=None):

    # Prepare
    if npts is None:
        npts = 1000
    angle = bragg + delta_bragg + 3.*sigma*np.linspace(-1, 1, npts)
    curve = func(angle)

    # Plot
    if ax is None:
        fig = plt.figure()
        ax = fig.add_axes([0.1, 0.1, 0.8, 0.8])
    ax.plot(angle, curve, ls='-', lw=1., c='k')
    ax.axvline(bragg, ls='--', lw=1, c='k')
    return ax


# #################################################################
# #################################################################
#                   Bragg diffraction plot
# #################################################################
# #################################################################

# Deprecated ? re-use ?
def CrystalBragg_plot_approx_detector_params(Rrow, bragg, d, Z,
                                             frame_cent, nn):

    R = 2.*Rrow
    L = 2.*R
    ang = np.linspace(0., 2.*np.pi, 100)

    fig = plt.figure()
    ax = fig.add_axes([0.1,0.1,0.8,0.8], aspect='equal')

    ax.axvline(0, ls='--', c='k')
    ax.plot(Rrow*np.cos(ang), Rrow + Rrow*np.sin(ang), c='r')
    ax.plot(R*np.cos(ang), R + R*np.sin(ang), c='b')
    ax.plot(L*np.cos(bragg)*np.r_[-1,0,1],
            L*np.sin(bragg)*np.r_[1,0,1], c='k')
    ax.plot([0, d*np.cos(bragg)], [Rrow, d*np.sin(bragg)], c='r')
    ax.plot([0, d*np.cos(bragg)], [Z, d*np.sin(bragg)], 'g')
    ax.plot([0, L/10*nn[1]], [Z, Z+L/10*nn[2]], c='g')
    ax.plot(frame_cent[1]*np.cos(2*bragg-np.pi),
            Z + frame_cent[1]*np.sin(2*bragg-np.pi), c='k', marker='o', ms=10)

    ax.set_xlabel(r'y')
    ax.set_ylabel(r'z')
    ax.legend(loc='upper left', bbox_to_anchor=(1.05, 1.), frameon=False)
    return ax


def CrystalBragg_plot_xixj_from_braggangle(bragg=None, xi=None, xj=None,
                                           data=None, ax=None):
    if ax is None:
        fig = plt.figure()
        ax = fig.add_axes([0.1,0.1,0.8,0.8], aspect='equal')

    for ii in range(len(bragg)):
        deg ='{0:07.3f}'.format(bragg[ii]*180/np.pi)
        ax.plot(xi[:,ii], xj[:,ii], '.', label='bragg %s'%deg)

    ax.set_xlabel(r'xi')
    ax.set_ylabel(r'yi')
    ax.legend(loc='upper left', bbox_to_anchor=(1.05, 1.), frameon=False)
    return ax




def CrystalBragg_plot_braggangle_from_xixj(xi=None, xj=None,
                                           bragg=None, angle=None,
                                           ax=None, plot=None,
                                           braggunits='rad', angunits='rad',
                                           **kwdargs):

    if isinstance(plot, bool):
        plot = 'contour'

    if ax is None:
        fig = plt.figure()
        ax0 = fig.add_axes([0.1, 0.1, 0.35, 0.8], aspect='equal')
        ax1 = fig.add_axes([0.55, 0.1, 0.35, 0.8], aspect='equal')
        ax = [ax0, ax1]
    if plot == 'contour':
        if 'levels' in kwdargs.keys():
            lvls = kwdargs['levels']
            del kwdargs['levels']
            obj0 = ax[0].contour(xi, xj, bragg, lvls, **kwdargs)
            obj1 = ax[1].contour(xi, xj, angle, lvls, **kwdargs)
        else:
            obj0 = ax[0].contour(xi, xj, bragg, **kwdargs)
            obj1 = ax[1].contour(xi, xj, angle, **kwdargs)
    elif plot == 'imshow':
        extent = (xi.min(), xi.max(), xj.min(), xj.max())
        obj0 = ax[0].imshow(bragg, extent=extent, aspect='equal',
                            adjustable='datalim', **kwdargs)
        obj1 = ax[1].imshow(angle, extent=extent, aspect='equal',
                            adjustable='datalim', **kwdargs)
    elif plot == 'pcolor':
        obj0 = ax[0].pcolor(xi, xj, bragg, **kwdargs)
        obj1 = ax[1].pcolor(xi, xj, angle, **kwdargs)
    ax[0].set_xlabel(r'xi')
    ax[1].set_xlabel(r'xi')
    ax[0].set_ylabel(r'yi')
    ax[1].set_ylabel(r'yi')
    cax0 = plt.colorbar(obj0, ax=ax[0])
    cax1 = plt.colorbar(obj1, ax=ax[1])
    cax0.ax.set_ylabel(r'$\theta_{bragg}$ (%s)'%braggunits)
    cax1.ax.set_ylabel(r'$ang$ (%s)'%angunits)
    return ax


def CrystalBragg_plot_line_tracing_on_det(lamb, xi, xj, xi_err, xj_err,
                                          det=None,
                                          johann=None, rocking=None,
                                          fs=None, dmargin=None,
                                          wintit=None, tit=None):

    # Check inputs
    # ------------

    if fs is None:
        fs = (6, 8)
    if dmargin is None:
        dmargin = {'left': 0.05, 'right': 0.99,
                   'bottom': 0.06, 'top': 0.92,
                   'wspace': None, 'hspace': 0.4}

    if wintit is None:
        wintit = _WINTIT
    if tit is None:
        tit = "line tracing"
        if johann is True:
            tit += " - johann error"
        if rocking is True:
            tit += " - rocking curve"

    plot_err = johann is True or rocking is True

    # Plot
    # ------------

    fig = fig = plt.figure(figsize=fs)
    gs = gridspec.GridSpec(1, 1, **dmargin)
    ax0 = fig.add_subplot(gs[0, 0], aspect='equal', adjustable='datalim')

    ax0.plot(det[0, :], det[1, :], ls='-', lw=1., c='k')
    for l in range(lamb.size):
        lab = r'$\lambda$'+' = {:6.3f} A'.format(lamb[l]*1.e10)
        l0, = ax0.plot(xi[l, :], xj[l, :], ls='-', lw=1., label=lab)
        if plot_err:
            ax0.plot(xi_err[l, ...], xj_err[l, ...],
                     ls='None', lw=1., c=l0.get_color(),
                     marker='.', ms=1)

    ax0.legend()

    if wintit is not False:
        fig.canvas.set_window_title(wintit)
    if tit is not False:
        fig.suptitle(tit, size=14, weight='bold')
    return [ax0]


def CrystalBragg_plot_johannerror(xi, xj, lamb, phi, err_lamb, err_phi,
                                  cmap=None, vmin=None, vmax=None,
                                  fs=None, dmargin=None, wintit=None, tit=None,
                                  angunits='deg', err=None):

    # Check inputs
    # ------------

    if fs is None:
        fs = (14, 8)
    if cmap is None:
        cmap = plt.cm.viridis
    if dmargin is None:
        dmargin = {'left': 0.05, 'right': 0.99,
                   'bottom': 0.06, 'top': 0.92,
                   'wspace': None, 'hspace': 0.4}
    assert angunits in ['deg', 'rad']
    if angunits == 'deg':
        # bragg = bragg*180./np.pi
        phi = phi*180./np.pi
        err_phi = err_phi*180./np.pi

    if err is None:
        err = 'abs'
    if err == 'rel':
        err_lamb = 100.*err_lamb / (np.nanmax(lamb) - np.nanmin(lamb))
        err_phi = 100.*err_phi / (np.nanmax(phi) - np.nanmin(phi))
        err_lamb_units = '%'
        err_phi_units = '%'
    else:
        err_lamb_units = 'm'
        err_phi_units = angunits

    if wintit is None:
        wintit = _WINTIT
    if tit is None:
        tit = False

    # pre-compute
    # ------------

    # extent
    extent = (xi.min(), xi.max(), xj.min(), xj.max())

    # Plot
    # ------------

    fig = fig = plt.figure(figsize=fs)
    gs = gridspec.GridSpec(1, 3, **dmargin)
    ax0 = fig.add_subplot(gs[0, 0], aspect='equal', adjustable='datalim')
    ax1 = fig.add_subplot(gs[0, 1], aspect='equal', adjustable='datalim',
                          sharex=ax0, sharey=ax0)
    ax2 = fig.add_subplot(gs[0, 2], aspect='equal', adjustable='datalim',
                          sharex=ax0, sharey=ax0)

    ax0.set_title('Iso-lamb and iso-phi at crystal summit')
    ax1.set_title('Focalization error on lamb ({})'.format(err_lamb_units))
    ax2.set_title('Focalization error on phi ({})'.format(err_phi_units))

    ax0.contour(xi, xj, lamb, 10, cmap=cmap)
    ax0.contour(xi, xj, phi, 10, cmap=cmap, ls='--')
    imlamb = ax1.imshow(err_lamb, extent=extent, aspect='equal',
                        origin='lower', interpolation='nearest',
                        vmin=vmin, vmax=vmax)
    imphi = ax2.imshow(err_phi, extent=extent, aspect='equal',
                       origin='lower', interpolation='nearest',
                       vmin=vmin, vmax=vmax)

    plt.colorbar(imlamb, ax=ax1)
    plt.colorbar(imphi, ax=ax2)
    if wintit is not False:
        fig.canvas.set_window_title(wintit)
    if tit is not False:
        fig.suptitle(tit, size=14, weight='bold')

    return [ax0, ax1, ax2]


def CrystalBragg_plot_data_vs_lambphi(xi, xj, bragg, lamb, phi, data,
                                      lambfit=None, phifit=None,
                                      spect1d=None, vertsum1d=None,
                                      lambax=None, phiax=None,
                                      cmap=None, vmin=None, vmax=None,
                                      fs=None, dmargin=None,
                                      angunits='deg'):

    # Check inputs
    # ------------

    if fs is None:
        fs = (14,8)
    if cmap is None:
        cmap = plt.cm.viridis
    if dmargin is None:
        dmargin = {'left':0.03, 'right':0.99,
                   'bottom':0.05, 'top':0.92,
                   'wspace':None, 'hspace':0.4}
    assert angunits in ['deg', 'rad']
    if angunits == 'deg':
        bragg = bragg*180./np.pi
        phi = phi*180./np.pi
        phifit = phifit*180./np.pi
        if phiax is not None:
            phiax = 180*phiax/np.pi



    # pre-compute
    # ------------

    # extent
    extent = (xi.min(), xi.max(), xj.min(), xj.max())
    extent2 = (lambfit.min(), lambfit.max(), phifit.min(), phifit.max())

    # Plot
    # ------------

    fig = fig = plt.figure(figsize=fs)
    gs = gridspec.GridSpec(4, 4, **dmargin)
    ax0 = fig.add_subplot(gs[:3, 0], aspect='equal', adjustable='datalim')
    ax1 = fig.add_subplot(gs[:3, 1], aspect='equal', adjustable='datalim',
                          sharex=ax0, sharey=ax0)
    axs1 = fig.add_subplot(gs[3, 1], sharex=ax0)
    ax2 = fig.add_subplot(gs[:3, 2])
    axs2 = fig.add_subplot(gs[3, 2], sharex=ax2, sharey=axs1)
    ax3 = fig.add_subplot(gs[:3, 3], sharey=ax2)

    ax0.set_title('Coordinates transform')
    ax1.set_title('Camera image')
    ax2.set_title('Camera image transformed')

    ax2.set_ylabel(r'incidence angle ($deg$)')
    ax2.set_xlabel(r'$\lambda$ ($m$)')
    axs2.set_xlabel(r'$\lambda$ ($m$)')
    ax3.set_ylabel(r'incidence angle ($deg$)')

    ax0.contour(xi, xj, bragg, 10, cmap=cmap)
    ax0.contour(xi, xj, phi, 10, cmap=cmap, ls='--')
    ax1.imshow(data, extent=extent, aspect='equal',
               origin='lower', vmin=vmin, vmax=vmax)
    axs1.plot(xi, np.nanmean(data, axis=0), c='k', ls='-')
    ax2.scatter(lamb.ravel(), phi.ravel(), c=data.ravel(), s=2,
                marker='s', edgecolors='None',
                cmap=cmap, vmin=vmin, vmax=vmax)
    axs2.plot(lambfit, spect1d, c='k', ls='-')
    ax3.plot(vertsum1d, phifit, c='k', ls='-')
    if phiax is not None:
        ax2.plot(lambax, phiax, c='r', ls='-', lw=1.)

    ax2.set_xlim(extent2[0], extent2[1])
    ax2.set_ylim(extent2[2], extent2[3])

    return [ax0, ax1]


def CrystalBragg_plot_data_vs_fit(xi, xj, bragg, lamb, phi, data, mask=None,
                                  lambfit=None, phifit=None, spect1d=None,
                                  dfit1d=None, dfit2d=None, lambfitbins=None,
                                  cmap=None, vmin=None, vmax=None,
                                  fs=None, dmargin=None,
                                  angunits='deg', dmoments=None):

    # Check inputs
    # ------------

    if fs is None:
        fs = (16, 9)
    if cmap is None:
        cmap = plt.cm.viridis
    if dmargin is None:
        dmargin = {'left':0.03, 'right':0.99,
                   'bottom':0.05, 'top':0.92,
                   'wspace':None, 'hspace':0.4}
    assert angunits in ['deg', 'rad']
    if angunits == 'deg':
        bragg = bragg*180./np.pi
        phi = phi*180./np.pi
        phifit = phifit*180./np.pi


    # pre-compute
    # ------------

    # extent
    extent = (xi.min(), xi.max(), xj.min(), xj.max())
    extent2 = (lambfit.min(), lambfit.max(), phifit.min(), phifit.max())

    ind = np.digitize(lamb[mask].ravel(), lambfitbins)
    spect2dmean = np.zeros((lambfitbins.size+1,))
    for ii in range(lambfitbins.size+1):
        indi = ind==ii
        if np.any(indi):
            spect2dmean[ii] = np.nanmean(dfit2d['fit'][indi])

    # Plot
    # ------------

    fig = plt.figure(figsize=fs)
    gs = gridspec.GridSpec(4, 6, **dmargin)
    ax0 = fig.add_subplot(gs[:3, 0], aspect='equal', adjustable='datalim')
    ax1 = fig.add_subplot(gs[:3, 1], aspect='equal', adjustable='datalim',
                          sharex=ax0, sharey=ax0)
    axs1 = fig.add_subplot(gs[3, 1], sharex=ax0)
    ax2 = fig.add_subplot(gs[:3, 2])
    axs2 = fig.add_subplot(gs[3, 2], sharex=ax2, sharey=axs1)
    ax3 = fig.add_subplot(gs[:3, 3], sharex=ax2, sharey=ax2)
    axs3 = fig.add_subplot(gs[3, 3], sharex=ax2)#, sharey=axs1)
    ax4 = fig.add_subplot(gs[:3, 4], sharex=ax2, sharey=ax2)
    axs4 = fig.add_subplot(gs[3, 3], sharex=ax2)#, sharey=axs1)
    ax5 = fig.add_subplot(gs[:3, 5], sharey=ax2)

    ax0.set_title('Coordinates transform')
    ax1.set_title('Camera image')
    ax2.set_title('Camera image transformed')
    ax3.set_title('2d spectral fit')
    ax4.set_title('2d error')
    ax5.set_title('Moments')

    ax4.set_xlabel('%s'%angunits)
    ax0.set_ylabel(r'incidence angle ($deg$)')

    ax0.contour(xi, xj, bragg, 10, cmap=cmap)
    ax0.contour(xi, xj, phi, 10, cmap=cmap, ls='--')
    ax1.imshow(data, extent=extent, aspect='equal',
               origin='lower', vmin=vmin, vmax=vmax)
    axs1.plot(xi, np.nanmean(data, axis=0), c='k', ls='-')
    ax2.scatter(lamb.ravel(), phi.ravel(), c=data.ravel(), s=1,
                marker='s', edgecolors='None',
                cmap=cmap, vmin=vmin, vmax=vmax)
    axs2.plot(lambfit, spect1d, c='k', ls='None', marker='.', ms=4)
    axs2.plot(lambfit, dfit1d['fit'].ravel(), c='r', ls='-', label='fit')
    for ll in dfit1d['lamb0']:
        axs2.axvline(ll, c='k', ls='--')

    # dfit2d
    ax3.scatter(lamb[mask].ravel(), phi[mask].ravel(), c=dfit2d['fit'], s=1,
                marker='s', edgecolors='None',
                cmap=cmap, vmin=vmin, vmax=vmax)
    axs3.plot(lambfit, spect1d, c='k', ls='None', marker='.')
    axs3.plot(lambfit, spect2dmean, c='b', ls='-')
    err = dfit2d['fit'] - data[mask].ravel()
    errmax = np.max(np.abs(err))
    ax4.scatter(lamb[mask].ravel(), phi[mask].ravel(), c=err, s=1,
                marker='s', edgecolors='None',
                cmap=plt.cm.seismic, vmin=-errmax, vmax=errmax)

    # Moments
    if dmoments is not None:
        if dmoments.get('ratio') is not None:
            ind = dmoments['ratio'].get('ind')
            if ind is None:
                ind = [np.argmin(np.abs(dfit2d['lamb0']-ll))
                        for ll in dmoments['ratio']['lamb']]
            for indi in ind:
                axs3.axvline(dfit2d['lamb0'][indi], c='k', ls='--')
            amp0 = BSpline(dfit2d['knots'], dfit2d['camp'][ind[0],:], dfit2d['deg'])(phifit)
            amp1 = BSpline(dfit2d['knots'], dfit2d['camp'][ind[1],:], dfit2d['deg'])(phifit)
            lab = dmoments['ratio']['name'] + '{} / {}'
            ratio = (amp0 / amp1) / np.nanmax(amp0 / amp1)
            ax5.plot(amp0 / amp1, phifit, ls='-', c='k', label=lab)
        if dmoments.get('sigma') is not None:
            ind = dmoments['sigma'].get('ind')
            if ind is None:
                ind = np.argmin(np.abs(dfit2d['lamb0']-dmoments['sigma']['lamb']))
            axs3.axvline(dfit2d['lamb0'][ind], c='b', ls='--')
            sigma = BSpline(dfit2d['knots'], dfit2d['csigma'][ind,:], dfit2d['deg'])(phifit)
            lab = r'$\sigma({} A)$'.format(np.round(dfit2d['lamb0'][ind]*1.e10),
                                        4)
            ax5.plot(sigma/np.nanmax(sigma), phifit, ls='-', c='b', label=lab)

    ax2.set_xlim(extent2[0], extent2[1])
    ax2.set_ylim(extent2[2], extent2[3])
    return [ax0, ax1]
